import sys
import inspect
import pandas as pd
from osier.technology import *
from osier.technology import _dim_opts
from unyt import GW, MW, kW, hour, day, year, kg
import unyt as u

to_MDOLLARS = 1e-6

co2_eq_units = (u.megatonnes*(GW*hour)**-1)

nuclear = ThermalTechnology(technology_name='Nuclear',
                            capacity=18.609404*GW,
                            capital_cost=50*(1/kW)*to_MDOLLARS,
                            om_cost_variable=0.0*to_MDOLLARS,
                            om_cost_fixed=177.73741*(1/kW)*to_MDOLLARS,
                            fuel_cost=5.811*(1/(MW*hour))*to_MDOLLARS,
                            ramp_up_rate=0.0,
                            ramp_down_rate=0.0,
                            lifecycle_co2_rate=5.1e-6*co2_eq_units,
                            land_intensity=0.0,
                            )
nuclear_adv = ThermalTechnology(technology_name='Nuclear_Adv',
                                capacity=0*GW,
                                capital_cost=4916.4*(1/kW)*to_MDOLLARS,
                                om_cost_variable=0.0*to_MDOLLARS,
                                om_cost_fixed=118.99*(1/kW)*to_MDOLLARS,
                                fuel_cost=9.158*(1/(MW*hour))*to_MDOLLARS,
                                ramp_up_rate=0.25,
                                ramp_down_rate=0.25,
                                lifecycle_co2_rate=nuclear.lifecycle_co2_rate,
                                land_intensity=0.0,
                                )
natural_gas = ThermalTechnology(technology_name='NaturalGas_Conv',
                                capacity=8.3751331*GW,
                                capital_cost=959.58*(1/kW)*to_MDOLLARS,
                                om_cost_variable=0.0*to_MDOLLARS,
                                om_cost_fixed=11.1934*(1/kW)*to_MDOLLARS,
                                fuel_cost=22.387*(1/(MW*hour))*to_MDOLLARS,
                                ramp_up_rate=1.0,
                                ramp_down_rate=1.0,
                                lifecycle_co2_rate=4.9e-4*co2_eq_units,
                                land_intensity=0.0,
                                )
biomass = ThermalTechnology(technology_name='Biomass',
                            renewable=True,
                            capacity=0*GW,
                            capital_cost=3436*(1/kW)*to_MDOLLARS,
                            om_cost_variable=0.0*to_MDOLLARS,
                            om_cost_fixed=123*(1/kW)*to_MDOLLARS,
                            fuel_cost=47*(1/(MW*hour))*to_MDOLLARS,
                            ramp_up_rate=1.0,
                            ramp_down_rate=1.0,
                            lifecycle_co2_rate=2.3e-4*co2_eq_units,
                            land_intensity=0.0,
                            )
natural_gas_adv = ThermalTechnology(technology_name='NaturalGas_Adv',
                                    capacity=0*GW,
                                    capital_cost=1891.0*(1/kW)*to_MDOLLARS,
                                    om_cost_variable=0.0*to_MDOLLARS,
                                    om_cost_fixed=26.99*(1/kW)*to_MDOLLARS,
                                    fuel_cost=27.475*(1/(MW*hour))*to_MDOLLARS,
                                    ramp_up_rate=1.0,
                                    ramp_down_rate=1.0,
                                    lifecycle_co2_rate=1.3e-4*co2_eq_units,
                                    land_intensity=0.0,
                                    )
coal = ThermalTechnology(technology_name='Coal_Conv',
                         capacity=0*GW,
                         capital_cost=1000*(1/kW)*to_MDOLLARS,
                         om_cost_variable=0.0*to_MDOLLARS,
                         om_cost_fixed=40.7033*(1/kW)*to_MDOLLARS,
                         fuel_cost=21.369*(1/(MW*hour))*to_MDOLLARS,
                         ramp_up_rate=0.5,
                         ramp_down_rate=0.5,
                         lifecycle_co2_rate=1e-3*co2_eq_units,
                         land_intensity=0.0,
                         )
coal_adv = ThermalTechnology(technology_name='Coal_Adv',
                            capacity=0*GW,
                            capital_cost=4924.6*(1/kW)*to_MDOLLARS,
                            om_cost_variable=0.0*to_MDOLLARS,
                            om_cost_fixed=58.24*(1/kW)*to_MDOLLARS,
                            fuel_cost=36.6329*(1/(MW*hour))*to_MDOLLARS,
                            ramp_up_rate=0.5,
                            ramp_down_rate=0.5,
                            lifecycle_co2_rate=3.7e-4*co2_eq_units,
                            land_intensity=0.0,
                            )
battery = StorageTechnology(technology_name='Battery',
                            capacity=0.81534126*GW,
                            capital_cost=613*(1/kW)*to_MDOLLARS,
                            om_cost_variable=0*to_MDOLLARS,
                            om_cost_fixed=15.32*(1/kW)*to_MDOLLARS,
                            fuel_cost=0*to_MDOLLARS,
                            storage_duration=4,
                            efficiency=0.85,
                            initial_storage=0,
                            capacity_credit=0.5,
                            lifecycle_co2_rate=3.3e-5*co2_eq_units,
                            land_intensity=0.0,
                            )
wind = Technology(technology_name='WindTurbine',
                  renewable=True,
                  dispatchable=False,
                  fuel_type='wind',
                  capacity=0*GW, 
                  capital_cost=1180.6*(1/kW)*to_MDOLLARS,
                  om_cost_fixed=33.11*(1/kW)*to_MDOLLARS,
                  lifecycle_co2_rate=1.2e-5*co2_eq_units,
                  land_intensity=0.0,
                  capacity_credit=0.35)
solar = Technology(technology_name='SolarPanel',
                  renewable=True,
                  dispatchable=False,
                  fuel_type='solar',
                  capacity=2.8103015*GW, 
                  capital_cost=673.2*(1/kW)*to_MDOLLARS,
                  om_cost_fixed=8.05*(1/kW)*to_MDOLLARS,
                  lifecycle_co2_rate=3.7e-5*co2_eq_units,
                  land_intensity=0.0,
                  capacity_credit=0.19)


def _get_names_technologies():
    """
    Returns a list of names and technologies in
    the :mod:`osier.tech_library`.
    """

    current_module = sys.modules[__name__]
    technology_list = []
    name_list = []

    for name, obj in inspect.getmembers(current_module):
        if isinstance(obj, Technology):
            technology_list.append(obj)
            name_list.append(name)
    
    return name_list, technology_list


def renewables_plus_storage():
    """
    Returns a list of technology objects including only
    renewable technologies and storage options.
    """

    names_list, technology_list = _get_names_technologies()
    technology_list = [tech
                       for tech in technology_list 
                       if ((tech.renewable) or (hasattr(tech, "storage_duration")))]

    return technology_list


def all_technologies():
    """
    Returns a list of all technology objects.
    """

    names_list, technology_list = _get_names_technologies()

    return technology_list


def catalog():
    """
    Returns a :class:`pandas.DataFrame` of technology names
    and their :class`osier` aliases.
    """

    names_list, technology_list = _get_names_technologies()
    catalog_df = pd.DataFrame({"Import Name":names_list,
                               "Technology Name":[t.technology_name 
                                                  for t in technology_list]})

    return catalog_df



import osier
from unyt import unyt_array
import unyt
import numpy as np
import pandas as pd
import pytest
import sys

TOL = 1e-5
N_HOURS = 24
N_DAYS = 2
N = N_HOURS * N_DAYS


@pytest.fixture
def net_demand():

    phase_shift = 0
    base_load = 1.5
    hours = np.linspace(0, N, N)
    demand = (np.sin((hours * np.pi / N_HOURS * 2 + phase_shift))
              * -1 + np.ones(N) * (base_load + 1))

    return demand

def test_deap_init(net_demand):
    """
    Tests that :class:`osier.models.deap_runner.OsierDEAP` is
    properly initialized.
    """
    techs = osier.all_technologies()
    objectives = [osier.total_cost, osier.annual_emission]
    problem = osier.CapacityExpansion(technology_list=techs,
                                      objectives=objectives,
                                      demand=net_demand,
                                      allow_blackout=False)
    model = osier.OsierDEAP(problem=problem,
                            pop_size=100)
    
    assert model.algorithm == 'nsga2'
    assert model.n_obj == 2
    assert model.completed_generations == 0
    assert model.n_dim == len(techs)
    assert model.solve_time == 0.0
    assert model.lower_bound == 0.0
    assert model.upper_bound == 1 / problem.capacity_credit.min()
    assert not model.repair
    assert not model.last_population
    assert not model.save_directory


from osier.utils import *
from unyt import kW, MW, minute, hour
import pytest


@pytest.fixture
def technology_set_1():
    """
    This fixture uses creates technologies directly from
    the :class:`Technology` class.
    """
    nuclear = Technology(technology_name='Nuclear',
                         technology_type='production',
                         capacity=5,
                         capital_cost=6,
                         om_cost_variable=20,
                         om_cost_fixed=50,
                         fuel_cost=5
                         )
    natural_gas = Technology(technology_name='NaturalGas',
                             technology_type='production',
                             capacity=5,
                             capital_cost=1,
                             om_cost_variable=12,
                             om_cost_fixed=30,
                             fuel_cost=20
                             )

    return [nuclear, natural_gas]


def test_synchronize_units(technology_set_1):
    """
    Tests that the synchronize units function makes
    all of the units in a list of technologies the
    same.
    """
    u_p = kW
    u_t = minute
    u_e = u_p * u_t

    synced = synchronize_units(technology_set_1, unit_power=u_p, unit_time=u_t)
    assert synced == technology_set_1
    assert [t.unit_power for t in synced] == [u_p, u_p]
    assert [t.unit_time for t in synced] == [u_t, u_t]
    assert [t.unit_energy for t in synced] == [u_e, u_e]



import unyt as u
from unyt import kg

try:
    u.define_unit("megatonnes", 1e9*kg)
except RuntimeError as e:
    print(e)
    u.unit_registry.default_unit_registry.remove("megatonnes")
    u.define_unit("megatonnes", 1e9*kg)


from .technology import *
from .models.dispatch import *
from .models.capacity_expansion import *
from .models.deap_runner import *
from .utils import *
from .equations import *

from .tech_library import *


import numpy as np
from pathlib import Path
import dill
import random
import time


from pymoo.core.problem import ElementwiseProblem, Problem
from pymoo.util.ref_dirs import get_reference_directions
from deap import algorithms
from deap import base
from deap.benchmarks.tools import igd
from deap import creator
from deap import tools

from osier import *
from osier.tech_library import *


def uniform(low, up, size=None):
    """
    Generates a random starting population from
    a uniform distribution.
    """
    try:
        return [random.uniform(a, b) for a, b in zip(low, up)]
    except TypeError:
        return [random.uniform(a, b) for a, b in zip([low] * size, [up] * size)]


algorithm_options = {'nsga2':tools.selNSGA2,
                     'nsga3':tools.selNSGA3}

class OsierDEAP(object):
    """
    A DEAP algorithm object. Holds information and parameters to
    run a genetic algorithm using DEAP.

    Parameters
    ----------
    problem : :class:`Osier.CapacityExpansion`
        The problem to execute. In theory, this could be any object
        that has a method :meth:`evaluate` (for example, if it inherits 
        from :class:`pymoo.ElementwiseProblem`) and outputs a dictionary.
    algorithm : Optional, str
        Indicates the algorithm to use. Accepts ['nsga2', 'nsga3']. 
        NSGA2 is recommended for problems with three or fewer objectives.
        NSGA3 is recommended for problems with three or more objectives. 
    lower_bound : Optional, float, int
        The lower bound for each 'gene' in a candidate individual.
    upper_bound : Optional, float, int
        The upper bound for each 'gene' in a candidate individual.
    repair : Optional, :class:`pymoo.Repair`
        Adjusts the candidate individuals based on a user-defined procedure.
    hyper_params : Optional, dict
        Contains hyperparameters for the selection algorithm.
    save_directory : Optional, str
        Indicates the read/write directory.
    """
    def __init__(self, 
                 problem,
                 algorithm=None, 
                 lower_bound=0.0,
                 upper_bound=None,
                 repair=None,
                 hyper_params=None,
                 pop_size=100,
                 save_directory=None
                 ) -> None:
        self.problem = problem
        self.n_obj = self.problem.n_obj
        if algorithm:
            self.algorithm = algorithm.lower()
        elif self.n_obj < 3:
            self.algorithm = 'nsga2'
        elif self.n_obj >= 3:
            self.algorithm = 'nsga3' 
        self.n_dim = self.problem.n_var
        self.pop_size = pop_size
        self.repair = repair
        self.completed_generations = 0
        self.solve_time = 0.0
        self.last_population = None
        self.save_directory = Path(save_directory) if save_directory else None

        try: 
            assert isinstance(problem, (ElementwiseProblem, Problem))
        except AssertionError as e:
            raise AssertionError(f"Problem type <{type(problem)}> is not supported.")
        self.lower_bound = lower_bound

        if upper_bound:
            self.upper_bound = upper_bound
        else:
            self.upper_bound = 1/self.problem.capacity_credit.min()

        # Algorithm parameters
        self.mutation_eta = hyper_params['mating.mutation.eta'] if hyper_params else 20.0
        self.mutation_prob = hyper_params['mating.mutation.prob'] if hyper_params else 1.0
        self.crossover_eta = hyper_params['mating.crossover.eta'] if hyper_params else 30.0
        self.crossover_prob = hyper_params['mating.crossover.prob'] if hyper_params else 1.0

        # DEAP setup        
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,) * self.n_obj)
        creator.create("Individual", list, fitness=creator.FitnessMin)
        
        self.toolbox = base.Toolbox()
        self.history = tools.History()
        self.logbook = tools.Logbook()
        self.logbook.header = "gen", "evals", "std", "min", "avg", "max"
        self.pareto_front = tools.ParetoFront()


        self.toolbox.register("attr_float", uniform, self.lower_bound, self.upper_bound, self.n_dim)
        self.toolbox.register("individual", tools.initIterate, creator.Individual, self.toolbox.attr_float)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)
        self.toolbox.register("evaluate", self.problem.evaluate, return_values_of=["F"])
        self.toolbox.register("mate", tools.cxSimulatedBinaryBounded, 
                              low=self.lower_bound, 
                              up=self.upper_bound, 
                              eta=self.crossover_eta)
        self.toolbox.register("mutate", tools.mutPolynomialBounded, 
                              low=self.lower_bound, 
                              up=self.upper_bound, 
                              eta=self.mutation_eta, 
                              indpb=1.0/self.n_dim)
        if self.algorithm == 'nsga3':
            self.ref_dirs = get_reference_directions('energy', self.n_obj, self.pop_size, seed=1)
            self.toolbox.register("select", algorithm_options[self.algorithm], ref_points=self.ref_dirs)
        elif self.algorithm == 'nsga2':
            self.ref_dirs = None
            self.toolbox.register("select", algorithm_options[self.algorithm])

        # Initialize statistics object
        self.stats = tools.Statistics(lambda ind: ind.fitness.values)
        self.stats.register("avg", np.mean, axis=0)
        self.stats.register("std", np.std, axis=0)
        self.stats.register("min", np.min, axis=0)
        self.stats.register("max", np.max, axis=0)


    def run(self, n_gen, seed=1234, init_pop=None, start_from_last=False):
        """
        Runs the genetic algorithm.

        Parameters
        ----------
        n_gen : int
            The number of generations to run.
        seed : int
            Sets the random seed.
        init_pop : :class:`deap.creator.Individual`
            Begins a simulation from a starting population.
        start_from_last : bool
            Begins a simulation from the last saved population.
        """
        random.seed(seed)

        if start_from_last and self.last_population:
            print('Starting from last population\n')
            pop = self.last_population
        elif init_pop:
            print('Starting from initialized population\n')
            pop = init_pop
        else:
            print('Starting from random population\n')
            pop = self.toolbox.population(n=self.pop_size)

        start = time.perf_counter()

        invalid_ind = [ind for ind in pop if not ind.fitness.valid]

        if self.repair and (len(invalid_ind) > 0):
            invalid_ind = self.repair._do(self.problem, np.array(invalid_ind))
            pop = [creator.Individual(ind) for ind in invalid_ind]
            invalid_ind = [ind for ind in pop if not ind.fitness.valid]

        try:
            fitnesses = self.toolbox.evaluate(np.array(invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit
        except AssertionError:
            print("Population already evaluated. Proceeding with genetic algorithm...")
        
        # Compile statistics about the population
        record = self.stats.compile(pop)
        self.logbook.record(gen=0, evals=len(pop), **record)
        self.history.update(pop)
        self.pareto_front.update(pop)
        self.last_population = pop
        print(self.logbook.stream)

        # Begin the generational process
        for gen in range(1, n_gen):
            offspring = algorithms.varAnd(pop, 
                                          self.toolbox, 
                                          self.crossover_prob, 
                                          self.mutation_prob)
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]

            # Evaluate the individuals with an invalid fitness
            if self.repair:
                invalid_ind = self.repair._do(self.problem, np.array(invalid_ind))
                offspring = [creator.Individual(ind) for ind in invalid_ind]
                invalid_ind = [ind for ind in offspring if not ind.fitness.valid]

            fitnesses = self.toolbox.evaluate(np.array(invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit

            # Select the next generation population from parents and offspring
            pop = self.toolbox.select(pop + offspring, self.pop_size)
            self.history.update(pop)
            self.pareto_front.update(pop)
            # Compile statistics about the new population
            record = self.stats.compile(pop)
            self.logbook.record(gen=gen, evals=len(invalid_ind), **record)
            self.last_population = pop
            self.completed_generations += 1
            print(self.logbook.stream)

        end = time.perf_counter()
        self.solve_time += (end-start)

        return self.last_population, self.logbook, self.pareto_front
    
    def save_model(self, fpath=None):
        """
        Serializes the model state in a binary file.
        """
        
        timestr = time.strftime("%Y%m%d-%H%M%S")
        save_name = f"{timestr}-OsierModel.pkl"
        if self.save_directory:
            fpath = self.save_directory / save_name
        else:
            fpath = save_name

        with open(fpath, "wb") as file:
            dill.dump(self.__dict__, file)
        

    def load_model(self, fpath):
        """
        Loads a serialized model into the current object.
        """

        with open(fpath, "rb") as file:
            tmp_dict = dill.load(file)
        
        self.__dict__.update(tmp_dict)


import pytest
import unyt
import numpy as np
import pandas as pd
import osier
from unyt import kW, MW, hr, BTU, Horsepower, day, kg, GW, megatonnes
from osier import Technology
from osier.technology import _validate_unit, _validate_quantity
from unyt.exceptions import UnitParseError

TECH_NAME = "PlanetExpress"
energy_unyt = 10.0 * MW * hr
other_energy_unyt = 10.0 * BTU
spec_energy_unyt = 10.0 * (MW * hr)**-1
power_unyt = 10.0 * MW
spec_power_unyt = 10.0 * MW**-1
other_power_unyt = Horsepower
time_unyt = 1 * hr
float_val = 10.0
int_val = 10
str_val = "10"
energy_str = "10 MW*hr"
mass_energy_str = "kg*BTU**-1"
spec_energy_str = "10 (MW*hr)**-1"
power_str = "10 MW"
spec_power_str = "10 MW**-1"
time_str = "10 hr"
unknown_str = "10 fortnights"
dict_type = {"value": 10,
             "unit": MW}
time_series_list = list(range(10))
time_series_np = np.array(time_series_list)
time_series_pd = pd.Series(time_series_list)
time_series_unyt = np.array(time_series_list) * time_unyt





@pytest.fixture
def advanced_tech():
    PLANET_EXPRESS = Technology(TECH_NAME)
    return PLANET_EXPRESS


def test_validate_unit():
    assert _validate_unit("MW", 'power').same_dimensions_as(Horsepower)
    assert _validate_unit("BTU", 'energy').same_dimensions_as(MW * hr)
    assert _validate_unit("day", 'time').same_dimensions_as(hr)
    assert _validate_unit(mass_energy_str, 'mass_per_energy').same_dimensions_as(kg*BTU**-1)
    assert _validate_unit(
        "Horsepower**-1",
        'specific_power').same_dimensions_as(
        MW**-1)
    assert _validate_unit(
        (Horsepower * day)**-1,
        'specific_energy').same_dimensions_as(
        (MW * hr)**-1)

    with pytest.raises(UnitParseError) as e:
        _validate_unit("darkmatter", "energy")

    with pytest.raises(KeyError) as e:
        _validate_unit("darkmatter", "fuel")

def test_validate_quantity():
    assert _validate_quantity(power_unyt, 'power') == 10 * (MW)
    assert _validate_quantity(energy_unyt, 'energy') == 10 * (MW * hr)
    assert _validate_quantity(time_unyt, 'time') == 1 * (hr)
    assert _validate_quantity("10 Horsepower**-1",
                              'specific_power') == 10 * (Horsepower**-1)
    assert _validate_quantity(10 * (Horsepower * day)**-1,
                              'specific_energy') == 10 * ((Horsepower * day)**-1)

    with pytest.raises(TypeError) as e:
        _validate_quantity(10 * MW, "energy")
    with pytest.raises(UnitParseError) as e:
        _validate_quantity("10 darkmatter", "energy")

    with pytest.raises(KeyError) as e:
        _validate_quantity("10 darkmatter", "fuel")

def test_validate_quantity_time_series():
    assert (_validate_quantity(time_series_list, 'time') == time_series_unyt).all()
    assert (_validate_quantity(time_series_np, 'time') == time_series_unyt).all()
    assert (_validate_quantity(time_series_pd, 'time') == time_series_unyt).all()
    assert (_validate_quantity(time_series_unyt, 'time') == time_series_unyt).all()


def test_initialize(advanced_tech):
    assert advanced_tech.technology_name == TECH_NAME
    assert advanced_tech.technology_type == 'production'
    assert advanced_tech.technology_category == 'base'
    assert advanced_tech.capacity == 0.0
    assert advanced_tech.capital_cost == 0.0
    assert advanced_tech.om_cost_fixed == 0.0
    assert advanced_tech.om_cost_variable == 0.0
    assert advanced_tech.fuel_cost == 0.0
    assert advanced_tech.unit_power == MW
    assert advanced_tech.unit_time == hr
    assert advanced_tech.unit_energy == MW * hr
    assert advanced_tech.unit_mass == megatonnes
    assert advanced_tech.annual_fixed_cost == 0.0
    assert advanced_tech.total_capital_cost == 0.0
    assert advanced_tech.efficiency == 1.0


def test_total_capital_cost(advanced_tech):
    advanced_tech.capital_cost = spec_power_unyt
    advanced_tech.capacity = power_unyt
    assert advanced_tech.total_capital_cost == 100

    advanced_tech.capacity = 0.5 * power_unyt
    assert advanced_tech.total_capital_cost == 50


def test_variable_cost(advanced_tech):
    advanced_tech.fuel_cost = spec_energy_unyt
    advanced_tech.om_cost_variable = spec_energy_unyt
    assert advanced_tech.variable_cost == 2 * spec_energy_unyt


def test_attribute_types(advanced_tech):
    assert isinstance(advanced_tech.capacity, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.capital_cost, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.om_cost_fixed, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.om_cost_variable, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.fuel_cost, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.co2_rate, unyt.array.unyt_quantity)
    assert isinstance(advanced_tech.unit_power, unyt.unit_object.Unit)
    assert isinstance(advanced_tech.unit_energy, unyt.unit_object.Unit)
    assert isinstance(advanced_tech.unit_time, unyt.unit_object.Unit)
    assert isinstance(advanced_tech.unit_mass, unyt.unit_object.Unit)


def test_capacity(advanced_tech):
    with pytest.raises(ValueError) as e:
        advanced_tech.capacity = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.capacity = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.capacity = energy_str

    advanced_tech.capacity = power_unyt
    assert advanced_tech.capacity.value == 10.0
    assert advanced_tech.capacity.units == MW

    advanced_tech.capacity = power_str
    assert advanced_tech.capacity.value == 10.0
    assert advanced_tech.capacity.units == MW

    advanced_tech.capacity = int_val
    assert advanced_tech.capacity.value == 10
    assert advanced_tech.capacity.units == MW

    advanced_tech.capacity = str_val
    assert advanced_tech.capacity.value == 10.0
    assert advanced_tech.capacity.units == MW

    advanced_tech.capacity = float_val * other_power_unyt
    assert advanced_tech.capacity.value == pytest.approx(0.007457, 0.005)
    assert advanced_tech.capacity.units == MW

    advanced_tech.unit_power = "kW"
    assert advanced_tech.capacity.units == kW


def test_capital_cost(advanced_tech):
    with pytest.raises(ValueError) as e:
        advanced_tech.capital_cost = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.capital_cost = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.capital_cost = energy_str

    advanced_tech.capital_cost = spec_power_unyt
    assert advanced_tech.capital_cost.value == 10.0
    assert advanced_tech.capital_cost.units == MW**-1

    advanced_tech.capital_cost = spec_power_str
    assert advanced_tech.capital_cost.value == 10.0
    assert advanced_tech.capital_cost.units == MW**-1

    advanced_tech.capital_cost = int_val
    assert advanced_tech.capital_cost.value == 10
    assert advanced_tech.capital_cost.units == MW**-1

    advanced_tech.capital_cost = str_val
    assert advanced_tech.capital_cost.value == 10.0
    assert advanced_tech.capital_cost.units == MW**-1

    advanced_tech.capital_cost = float_val / other_power_unyt
    assert advanced_tech.capital_cost.value == pytest.approx(13410.220, 0.005)
    assert advanced_tech.capital_cost.units == MW**-1

    advanced_tech.unit_power = "kW"
    assert advanced_tech.capital_cost.units == (kW)**-1


def test_om_cost_fixed(advanced_tech):
    with pytest.raises(ValueError) as e:
        advanced_tech.om_cost_fixed = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.om_cost_fixed = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.om_cost_fixed = energy_str

    advanced_tech.om_cost_fixed = spec_power_unyt
    assert advanced_tech.om_cost_fixed.value == 10.0
    assert advanced_tech.om_cost_fixed.units == MW**-1

    advanced_tech.om_cost_fixed = spec_power_str
    assert advanced_tech.om_cost_fixed.value == 10.0
    assert advanced_tech.om_cost_fixed.units == MW**-1

    advanced_tech.om_cost_fixed = int_val
    assert advanced_tech.om_cost_fixed.value == 10
    assert advanced_tech.om_cost_fixed.units == MW**-1

    advanced_tech.om_cost_fixed = str_val
    assert advanced_tech.om_cost_fixed.value == 10.0
    assert advanced_tech.om_cost_fixed.units == MW**-1

    advanced_tech.om_cost_fixed = float_val / other_power_unyt
    assert advanced_tech.om_cost_fixed.value == pytest.approx(13410.220, 0.005)
    assert advanced_tech.om_cost_fixed.units == MW**-1

    advanced_tech.unit_power = "kW"
    assert advanced_tech.om_cost_fixed.units == (kW)**-1


def test_om_cost_variable(advanced_tech):
    with pytest.raises(ValueError) as e:
        advanced_tech.om_cost_variable = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.om_cost_variable = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.om_cost_variable = power_str
    with pytest.raises(ValueError) as e:
        advanced_tech.om_cost_variable = spec_energy_str
    assert advanced_tech.om_cost_variable.value == 0.0
    assert advanced_tech.om_cost_variable.units == (MW*hr)**-1

    advanced_tech.om_cost_variable = spec_energy_unyt
    assert advanced_tech.om_cost_variable.value == 10.0
    assert advanced_tech.om_cost_variable.units == (MW * hr)**-1

    advanced_tech.om_cost_variable = int_val
    assert advanced_tech.om_cost_variable.value == 10
    assert advanced_tech.om_cost_variable.units == (MW * hr)**-1

    advanced_tech.om_cost_variable = str_val
    assert advanced_tech.om_cost_variable.value == 10.0
    assert advanced_tech.om_cost_variable.units == (MW * hr)**-1

    advanced_tech.om_cost_variable = float_val / other_energy_unyt
    assert advanced_tech.om_cost_variable.value == pytest.approx(
        3412141.5, 0.5)
    assert advanced_tech.om_cost_variable.units == (MW * hr)**-1

    advanced_tech.unit_power = "kW"
    advanced_tech.unit_time = "day"
    assert advanced_tech.om_cost_variable.units == (kW * day)**-1


def test_fuel_cost(advanced_tech):
    with pytest.raises(ValueError) as e:
        advanced_tech.fuel_cost = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.fuel_cost = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.fuel_cost = power_str
    with pytest.raises(ValueError) as e:
        advanced_tech.fuel_cost = spec_energy_str

    advanced_tech.fuel_cost = spec_energy_unyt
    assert advanced_tech.fuel_cost.value == 10.0
    assert advanced_tech.fuel_cost.units == (MW * hr)**-1

    advanced_tech.fuel_cost = int_val
    assert advanced_tech.fuel_cost.value == 10
    assert advanced_tech.fuel_cost.units == (MW * hr)**-1

    advanced_tech.fuel_cost = str_val
    assert advanced_tech.fuel_cost.value == 10.0
    assert advanced_tech.fuel_cost.units == (MW * hr)**-1

    advanced_tech.fuel_cost = float_val / other_energy_unyt
    assert advanced_tech.fuel_cost.value == pytest.approx(
        3412141.47989694, 0.5)
    assert advanced_tech.fuel_cost.units == (MW * hr)**-1

    advanced_tech.unit_power = "kW"
    advanced_tech.unit_time = "day"
    assert advanced_tech.fuel_cost.units == (kW * day)**-1


def test_co2_rate(advanced_tech):
    expected_unit = megatonnes*(MW*hr)**-1
    with pytest.raises(ValueError) as e:
        advanced_tech.co2_rate = dict_type
    with pytest.raises(UnitParseError) as e:
        advanced_tech.co2_rate = unknown_str
    with pytest.raises(AssertionError) as e:
        advanced_tech.co2_rate = power_str
    with pytest.raises(ValueError) as e:
        advanced_tech.co2_rate = spec_energy_str
    assert advanced_tech.co2_rate.value == 0.0
    assert advanced_tech.co2_rate.units == expected_unit

    advanced_tech.co2_rate = megatonnes*spec_energy_unyt
    assert advanced_tech.co2_rate.value == 10.0
    assert advanced_tech.co2_rate.units == expected_unit

    advanced_tech.co2_rate = int_val
    assert advanced_tech.co2_rate.value == 10
    assert advanced_tech.co2_rate.units == expected_unit

    advanced_tech.co2_rate = str_val
    assert advanced_tech.co2_rate.value == 10.0
    assert advanced_tech.co2_rate.units == expected_unit

    advanced_tech.co2_rate = float_val * megatonnes / other_energy_unyt
    assert advanced_tech.co2_rate.value == pytest.approx(
        3412141.5, 0.5)
    assert advanced_tech.co2_rate.units == expected_unit

    advanced_tech.unit_power = "kW"
    advanced_tech.unit_time = "day"
    assert advanced_tech.co2_rate.units == megatonnes*(kW * day)**-1

    advanced_tech.unit_power = "GW"
    advanced_tech.unit_time = "hr"
    advanced_tech.unit_mass = "ton"
    advanced_tech.co2_rate = 1e-2 * megatonnes*(GW*hr)**-1
    assert advanced_tech.co2_rate == pytest.approx(11023.113, 0.1)


def test_unit_power(advanced_tech):
    with pytest.raises(UnitParseError) as e:
        advanced_tech.unit_power = "darkmatter"
    with pytest.raises(AssertionError) as e:
        advanced_tech.unit_power = BTU
    with pytest.raises(AssertionError) as e:
        advanced_tech.unit_power = "BTU"
    with pytest.raises(ValueError) as e:
        advanced_tech.unit_power = 10
    advanced_tech.unit_power = Horsepower
    assert advanced_tech.unit_power == Horsepower

    advanced_tech.unit_power = "Horsepower"
    assert advanced_tech.unit_power == Horsepower


def test_unit_time(advanced_tech):
    with pytest.raises(UnitParseError) as e:
        advanced_tech.unit_time = "darkmatter"
    with pytest.raises(AssertionError) as e:
        advanced_tech.unit_time = MW
    with pytest.raises(AssertionError) as e:
        advanced_tech.unit_time = "MW"
    with pytest.raises(ValueError) as e:
        advanced_tech.unit_time = 10
    advanced_tech.unit_time = day
    assert advanced_tech.unit_time == day

    advanced_tech.unit_time = "day"
    assert advanced_tech.unit_time == day


def test_unit_energy(advanced_tech):
    advanced_tech.unit_energy = "darkmatter"
    assert advanced_tech.unit_energy == MW * hr
    advanced_tech.unit_energy = BTU
    assert advanced_tech.unit_energy == MW * hr
    advanced_tech.unit_energy = "MW"
    assert advanced_tech.unit_energy == MW * hr
    advanced_tech.unit_energy = 10
    assert advanced_tech.unit_energy == MW * hr
    advanced_tech.unit_energy = Horsepower * day
    assert advanced_tech.unit_energy == MW * hr
    advanced_tech.unit_energy = "Horsepower*day"
    assert advanced_tech.unit_energy == MW * hr



from osier import Technology
from osier.technology import _validate_unit
from unyt import unit_object
import copy
from typing import Iterable
import pandas as pd

from osier.technology import Technology


def synchronize_units(tech_list: Iterable[Technology],
                      unit_power: unit_object,
                      unit_time: unit_object) -> Iterable[Technology]:
    """
    This function ensures that all objects in the technology list
    have units consistent with the model's units. An
    :class:`osier.Technology` object (or sub-classes) have three
    unit settings.

        * :attr:`unit_power`
        * :attr:`unit_time`
        * :attr:`unit_energy`

    Only the :attr:`unit_power` and :attr:`unit_time` attributes can
    be specified with this function to prevent inconsistent units.
    E.g. Setting the :attr:`unit_energy` to ``MWh`` even though time
    is in minutes and power is in ``kW``.

    .. note::
        The objects in the original list are copied by default. This
        may result in slower run time.

    Parameters
    ----------
    tech_list : list of :class:`osier.Technology` objects
        The list of technology objects whose units need to be
        synchronized.

    Returns
    -------
    synced_list : list of :class:`osier.Technology` objects
        The list of technology objects that have synchronized units.
    """

    synced_list = [copy.deepcopy(t) for t in tech_list]

    for t in synced_list:
        t.unit_power = _validate_unit(unit_power, dimension="power")
        t.unit_time = _validate_unit(unit_time, dimension="time")

    return synced_list



def get_tech_names(technology_list):
    """
    Returns the a list of :class:`osier.Technology` name strings.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.

    Returns
    -------
    tech_names : list of str
        The list of technology names.
    """

    tech_names = [t.technology_name for t in technology_list]

    return tech_names


def get_dispatchable_techs(technology_list):
    """
    Returns a list of :class:`osier.Technology` objects 
    where :attr:`dispatchable` is `True`.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.

    Returns
    -------
    tech_names : list of :class:`osier.Technology`
        The list of dispatchable technologies.
    """

    dispatchable_techs = [t for t in technology_list if t.dispatchable]

    return dispatchable_techs


def get_nondispatchable_techs(technology_list):
    """
    Returns a list of :class:`osier.Technology` objects 
    where :attr:`dispatchable` is `False`.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.

    Returns
    -------
    non_dispatchable_techs : list of :class:`osier.Technology`
        The list of non dispatchable technologies.
    """

    non_dispatchable_techs = [t for t in technology_list if not t.dispatchable]

    return non_dispatchable_techs


def get_dispatchable_names(technology_list):
    """
    Returns a list of :class:`osier.Technology` name strings
    where :attr:`dispatchable` is `True`.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.

    Returns
    -------
    dispatchable_names : list of str
        The list of dispatchable technology names.
    """

    dispatchable_names = [
        t.technology_name for t in technology_list if t.dispatchable]

    return dispatchable_names


def technology_dataframe(technology_list, cast_to_string=True):
    """
    Returns a :class:`pandas.DataFrame` with a complete set
    of data for a given technology list.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.

    Returns
    -------
    technology_dataframe : :class:`pandas.DataFrame`
        A dataframe of all technology data.
    """

    frames = []

    for t in technology_list:
        frames.append(t.to_dataframe(cast_to_string=cast_to_string))

    technology_dataframe = pd.concat(frames, axis=0)

    return technology_dataframe
    




from osier.utils import (get_tech_names, 
                         get_dispatchable_techs, 
                         get_nondispatchable_techs, get_dispatchable_names)
from pyentrp.entropy import weighted_permutation_entropy
from functools import partial
import numpy as np
from unyt import unyt_array


def objective_from_capacity(technology_list,
                            attribute,
                            solved_dispatch_model=None):
    """
    This function calculates a general objective for a given
    set of technologies and their corresponding dispatch on a
    per-unit-capacity basis.

    The general objective function is

    .. math::
        \\mathcal{K} = \\sum_g^G \\textbf{CAP}_g \\kappa_g,

    .. math::
        \\textbf{CAP}_g = \\text{the capacity of the g-th technology}
        \\quad \\left[MW\\right].
        
    .. math::
        \\kappa_g = \\text{the power density of the g-th technology} 
        \\quad \\left[\\frac{-}{MW}\\right].

    .. warning::
        User-defined attributes are not validated by :class:`osier`.
        Verify the units are accurate and uniform across all
        technologies before running a simulation with this function.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).
    attribute : string
        The technology attribute to measure.

    Returns
    -------
    objective_value : float
        The objective value for a particular energy mix.

    Examples
    --------
    The simplest way to employ this function is with `functools.partial`.

    >>> import functools
    >>> from osier import per_unit_capacity
    >>> objectives_list = [functools.partial(objective_from_capacity, attribute='land_use'),
    ...                    functools.partial(objective_from_capacity, attribute='employment')]

    """
    objective_value = np.array([getattr(t, attribute) * t.capacity
                                for t in technology_list
                                if hasattr(t, attribute)]).sum()

    return objective_value


def objective_from_energy(technology_list, attribute, solved_dispatch_model):
    """
    This function calculates a general objective for a given
    set of technologies and their corresponding dispatch on a
    per-unit-energy basis.

    The general objective function is 
     
    .. math::
        \\mathcal{X} = \\sum_g^G \\xi_g \\sum_t^T x_{g,t},
    
    .. math::
        x_{g,t} = \\text{the energy produced by the g-th technology
         at time t}\\quad \\left[MWh\\right]

    .. math::
        \\xi_g = \\text{the energy density of the g-th technology}\\quad
        \\left[\\frac{-}{MWh}\\right].

        
    .. warning::
        User-defined attributes are not validated by :class:`osier`.
        Verify the units are accurate and uniform across all
        technologies before running a simulation with this function.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).
    attribute : string
        The technology attribute to measure.

    Returns
    -------
    objective_value : float
        The objective value for a particular energy mix.

    Examples
    --------
    The simplest way to employ this function is with `functools.partial`.

    >>> import functools
    >>> from osier import per_unit_capacity
    >>> objectives_list = [functools.partial(objective_from_energy, attribute='water_use'),
    ...                    functools.partial(objective_from_energy, attribute='death_rate')]

    """

    valid_techs = [t for t in technology_list if hasattr(t, attribute)]
    column_names = get_tech_names(valid_techs)
    dispatch_results = solved_dispatch_model.results
    power_units = solved_dispatch_model.power_units
    time_delta = solved_dispatch_model.time_delta
    dispatch_values = dispatch_results[column_names].values.T*power_units*time_delta
    
    mass_u = valid_techs[0].unit_mass
    attributes = unyt_array([getattr(t, attribute)
                           for t in valid_techs])
    
    attributes = attributes.to(mass_u*(power_units*time_delta.units)**-1)*time_delta.to_value()

    objective_value = np.dot(
        attributes,
        dispatch_values).sum()
    

    return objective_value.to_value()


def annualized_capital_cost(technology_list, solved_dispatch_model=None):
    """
    This function calculates the annual capital cost for a given
    set of technologies and their corresponding dispatch.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).

    Returns
    -------
    capital_cost : float
        The annual capital cost of the technology set.
    """
    capital_cost = np.array([t.total_capital_cost / t.lifetime
                            for t in technology_list]).sum()

    return capital_cost


def annualized_fixed_cost(technology_list, solved_dispatch_model=None):
    """
    This function calculates the annual fixed cost for a given
    set of technologies.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).

    Returns
    -------
    fixed_cost : float
        The annual fixed cost of the technology set.
    """
    fixed_cost = np.array([t.annual_fixed_cost
                           for t in technology_list]).sum()

    return fixed_cost


def annual_emission(
        technology_list,
        solved_dispatch_model,
        emission='lifecycle_co2_rate'):
    """
    This function calculates the total system co2 emissions for a given
    set of technologies and their corresponding dispatch.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).

    Returns
    -------
    emissions_total : float
        The total emissions of the technology set.
    """

    emissions_total = objective_from_energy(technology_list=technology_list, 
                                            solved_dispatch_model=solved_dispatch_model, 
                                            attribute=emission)

    return emissions_total

annual_co2 = partial(annual_emission, attribute='co2')  # for backwards compatibility

def total_cost(technology_list, solved_dispatch_model):
    """
    This function calculates the total system cost for a given
    set of technologies and their corresponding dispatch.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).

    Returns
    -------
    total_cost : float
        The total system cost.
    """

    capital_cost = annualized_capital_cost(technology_list)
    fixed_cost = annualized_fixed_cost(technology_list)
    variable_cost = solved_dispatch_model.objective
    total_cost = capital_cost + fixed_cost + variable_cost

    return total_cost


def volatility(technology_list,
               solved_dispatch_model,
               attribute='demand',
               m=3,
               tau=60,
               normalize=True):
    """
    This function calculates the volatility of the electricity prices
    or the electricity demand for a dispatch model with weighted 
    permutation entropy.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        The list of technologies.
    solved_dispatch_model : :class:`osier.DispatchModel`
        A _solved_ dispatch model (i.e. with model results and objective
        attributes).
    attribute : {'demand', 'volatility'}
        Indicates whether to calculate the volatility of electricity demand
        or electricity price.
    m : int
        The embedding dimension for the cost time series. Typically
        determined using a false nearest neighbors algorithm.
        The default value is 3.
    tau : int
        The time delay for the cost time series. Typically determined
        by selecting the index of the first or second minimum of the
        time series' delayed mutual information.

    Returns
    -------
    wpe : float
        The weighted permutation entropy of the cost time series.

    Notes
    -----
    Users can modify the parameters for this function using :attr:`functools.partial`.

    >>> import functools
    >>> from osier import volatility
    >>> objectives_list = [functools.partial(volatility, m=4, tau=100)]
    """

    timeseries = {'price':solved_dispatch_model.results.Cost.values,
                  'demand':np.array(solved_dispatch_model.net_demand)}
    wpe = weighted_permutation_entropy(timeseries[attribute],
                                       order=m,
                                       delay=tau,
                                       normalize=normalize)

    return wpe



import numpy as np
import pandas as pd
from copy import deepcopy
import dill
import unyt as u
from unyt import unyt_array
import functools
import time

from osier import DispatchModel

from pymoo.core.problem import ElementwiseProblem

LARGE_NUMBER = 1e20


class CapacityExpansion(ElementwiseProblem):
    """
    The :class:`CapacityExpansion` class inherits from the
    :class:`pymoo.core.problem.ElementwiseProblem` class. This problem
    determines the technology mix that _minimizes_ the provided objectives.

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology` objects
        Defines the technologies used in the model and the number of decision
        variables.
    demand : :class:`numpy.ndarray`
        The demand curve that needs to be met by the technology mix.
    objectives : list of str or functions
        Specifies the number and type of objectives. A list of strings must
        correspond to preset objective functions. Users may optionally write
        their own functions and pass them to `osier` as items in the list.
    constraints : dictionary of string : float or function : float
        Specifies the number and type of constraints. String key names must
        correspond to preset constraints functions. Users may optionally write
        their own functions and pass them to `osier` as keys in the list. The
        values must be numerical and represent the value that the function
        should not exceed. See notes for more information about constraints.
    prm : Optional, float
        The "planning reserve margin" (`prm`) specifies the amount of excess
        capacity needed to meet reliability standards. See
        :attr:`capacity_requirement`. Default is 0.0.
    solar : Optional, :class:`numpy.ndarray`
        The curve that defines the solar power provided at each time step.
        Automatically normalized with the infinity norm (i.e. divided by the
        maximum value).
    wind : Optional, :class:`numpy.ndarray`
        The curve that defines the wind power provided at each time step.
        Automatically normalized with the infinity norm (i.e. divided by the
        maximum value).
    power_units : str, :class:`unyt.unit_object`
        Specifies the units for the power demand. The default is :attr:`MW`. Can
        be overridden by specifying a unit with the value.
    penalty : Optional, float
        The penalty for infeasible solutions. If a particular set produces an
        infeasible solution for the :class:`osier.DispatchModel`, the
        corresponding objectives take on this value.
    curtailment : boolean
        Indicates if the model should enable a curtailment option.
    allow_blackout : boolean
        If True, a "reliability" technology is added to the dispatch model that
        will fulfill the mismatch in supply and demand. This reliability
        technology has a variable cost of 1e4 $/MWh. The value must be higher
        than the variable cost of any other technology to prevent a pathological
        preference for blackouts. Default is False.
    verbosity : Optional, int
        Sets the logging level for the simulation. Accepts `logging.LEVEL` or
        integer where LEVEL is {10:DEBUG, 20:INFO, 30:WARNING, 40:ERROR,
        50:CRITICAL}.
    solver : str
        Indicates which solver to use. May require separate installation.
        Accepts: ['cplex', 'cbc', 'glpk']. Other solvers will be added in the
        future.

    Notes
    -----
    **Constraints**:

    `Pymoo` constraints are not strict in the sense that `Pymoo` prefers
    feasibility over respecting constraints. However, all `Pymoo` algorithms
    will minimize the "constraint violation (CV)."
    """

    def __init__(self,
                 technology_list,
                 demand,
                 objectives,
                 constraints={},
                 solar=None,
                 wind=None,
                 prm=0.0,
                 lower_bound=0.0,
                 upper_bound=1.0,
                 penalty=LARGE_NUMBER,
                 power_units=u.MW,
                 curtailment=True,
                 allow_blackout=False,
                 verbosity=50,
                 solver='cbc',
                 **kwargs):
        self.technology_list = deepcopy(technology_list)
        self.demand = demand
        self.prm = prm

        self.objectives = objectives
        self.constraints = constraints
        self.penalty = penalty
        self.curtailment = curtailment
        self.allow_blackout = allow_blackout
        self.verbosity = verbosity
        self.solver = solver

        if isinstance(demand, unyt_array):
            self.power_units = demand.units
        else:
            self.power_units = power_units

        self.max_demand = float(demand.max()) * self.power_units
        self.capacity_requirement = self.max_demand * (1 + self.prm)

        if solar is not None:
            self.solar_ts = solar / solar.max()
        else:
            self.solar_ts = np.zeros(len(self.demand))
        if wind is not None:
            self.wind_ts = wind / wind.max()
        else:
            self.wind_ts = np.zeros(len(self.demand))

        super().__init__(n_var=len(self.technology_list),
                         n_obj=len(self.objectives),
                         n_constr=len(self.constraints),
                         xl=lower_bound,
                         xu=upper_bound,
                         **kwargs)

    def print_problem_formulation(self):
        """
        Prints the problem formulation.
        """

        print("===========================")
        print("CapacityExpansion Problem")
        print("===========================")
        print("Technologies:\n")
        print("Technology Name | Capacity \n")
        for t in self.technology_list:
            print(t)

        print("\nElectricity Demand:\n")
        print(self.demand)

        return

    @property
    def capacity_credit(self):
        return np.array([t.capacity_credit for t in self.technology_list])

    @property
    def dispatchable_techs(self):
        return [t for t in self.technology_list if t.dispatchable]

    def _evaluate(self, x, out, *args, **kwargs):
        capacities = self.capacity_requirement * x
        solar_capacity = 0
        wind_capacity = 0
        firm_capacity = 0
        for capacity, tech in zip(capacities, self.technology_list):
            tech.capacity = capacity
            if tech.renewable:
                if tech.fuel_type == 'solar':
                    solar_capacity += capacity
                elif tech.fuel_type == 'wind':
                    wind_capacity += capacity
            elif ((tech.dispatchable) and (tech.technology_name != 'Battery')):
                firm_capacity += capacity

        solar_gen = self.solar_ts * solar_capacity
        wind_gen = self.wind_ts * wind_capacity

        renewable_df = pd.DataFrame({'SolarPanel': solar_gen,
                                     'WindTurbine': wind_gen})

        net_demand = self.demand \
            - wind_gen \
            - solar_gen

        model = DispatchModel(technology_list=self.dispatchable_techs,
                              net_demand=net_demand,
                              power_units=self.power_units,
                              curtailment=self.curtailment,
                              allow_blackout=self.allow_blackout,
                              solver=self.solver,
                              verbosity=self.verbosity)
        model.solve()

        if model.results is not None:

            model.results = pd.concat([model.results, renewable_df], axis=1)

            out_obj = []
            for obj_func in self.objectives:
                out_obj.append(obj_func(technology_list=self.technology_list,
                                        solved_dispatch_model=model))

            if self.n_constr > 0:
                out_constr = []
                for constr_func, val in self.constraints.items():
                    out_constr.append(
                        constr_func(
                            technology_list=self.technology_list,
                            solved_dispatch_model=model) - val)

        else:
            out_obj = np.ones(self.n_obj) * self.penalty
            if self.n_constr > 0:
                out_constr = np.ones(self.n_constr) * self.penalty

        out["F"] = out_obj

        if self.n_constr > 0:
            out["G"] = out_constr



import unyt
from unyt import MW, hr, kg, km, m, megatonnes
from unyt import unyt_quantity, unyt_array
from unyt.exceptions import UnitParseError
from collections import OrderedDict

import numpy as np
import pandas as pd


_dim_opts = {'time': hr,
             'power': MW,
             'energy': MW * hr,
             'mass': kg,
             'length': km,
             'area': km**2,
             'volume': m**3,
             'specific_time': hr**-1,
             'specific_mass': kg**-1,
             'specific_power': MW**-1,
             'specific_energy': (MW * hr)**-1,
             'mass_per_energy': megatonnes * (MW * hr)**-1,
             'area_per_power': km * MW**-2}

_constant_types = (int, float, unyt_quantity)
_array_types = (unyt.unyt_array, pd.core.series.Series, np.ndarray, list)


def _validate_unit(value, dimension):
    """
    This function checks that a unit has the correct
    dimensions. Used in :class:`Technology` to set
    units.

    Parameters
    ----------
    value : string, float, int, or :class:`unyt.unit_object.Unit`
        The value being tested. Should be a unit symbol.
    dimension : string
        The expected dimensions of `value`.
        Accepted values listed in `_dim_opts`.

    Returns
    -------
    valid_unit : :class:`unyt.unit_object.Unit`
        The validated unit.
    """
    try:
        exp_dim = _dim_opts[dimension]
    except KeyError:
        raise KeyError(f"Key <{dimension}> not accepted. Try: {_dim_opts}")

    valid_unit = None
    if isinstance(value, unyt.unit_object.Unit):
        assert value.same_dimensions_as(exp_dim)
        valid_unit = value
    elif isinstance(value, str):
        try:
            unit = unyt_quantity.from_string(value).units
            assert unit.same_dimensions_as(exp_dim)
            valid_unit = unit
        except UnitParseError:
            raise UnitParseError(f"Could not interpret <{value}>.")
        except AssertionError:
            raise AssertionError(f"{value} lacks units of {dimension}.")
    else:
        raise ValueError(f"Value of type <{type(value)}> passed.")

    return valid_unit


def _validate_quantity(value, dimension):
    """
    This function checks that a quantity has the correct
    dimensions. Used in :class:`Technology` to set
    data attributess.

    Parameters
    ----------
    value : string, float, int, or :class:`unyt.unyt_quantity`
        The value being tested. Should be something like

        >>> _validate_quantity("10 MW", dimension='power')
        unyt_quantity(10., 'MW')

    dimension : string
        The expected dimensions of `value`.
        Accepted values listed in `_dim_opts`.

    Returns
    -------
    valid_quantity : :class:`unyt.unyt_quantity`
        The validated quantity.
    """
    try:
        exp_dim = _dim_opts[dimension]
    except KeyError:
        raise KeyError(f"Key <{dimension}> not accepted. Try: {_dim_opts}")

    valid_quantity = None
    if isinstance(value, unyt_quantity):
        try:
            assert value.units.same_dimensions_as(exp_dim)
            valid_quantity = value
        except AssertionError:
            raise TypeError(
                f"{value} has dimensions {value.units.dimensions}. "
                f"Expected {exp_dim.dimensions}")
    elif isinstance(value, unyt_array):
        try:
            assert value.units.same_dimensions_as(exp_dim)
            valid_quantity = value
        except AssertionError:
            raise TypeError(
                f"{value} has dimensions {value.units.dimensions}. "
                f"Expected {exp_dim.dimensions}")
    elif isinstance(value, np.ndarray):
        valid_quantity = value * exp_dim
    elif isinstance(value, pd.core.series.Series):
        valid_quantity = value.values * exp_dim
    elif isinstance(value, list):
        valid_quantity = np.array(value) * exp_dim
    elif isinstance(value, float):
        valid_quantity = value * exp_dim
    elif isinstance(value, int):
        valid_quantity = value * exp_dim
    elif isinstance(value, str):
        try:
            valid_quantity = float(value) * exp_dim
        except ValueError:
            try:
                unyt_value = unyt_quantity.from_string(value)
                assert unyt_value.units.same_dimensions_as(exp_dim)
                valid_quantity = unyt_value
            except UnitParseError:
                raise UnitParseError(f"Could not interpret <{value}>.")
            except AssertionError:
                raise AssertionError(f"{value} lacks units of {dimension}.")
    else:
        raise ValueError(f"Value of type <{type(value)}> passed.")
    return valid_quantity


class Technology(object):
    """
    The :class:`Technology` base class contains the minimum required
    data to solve an energy systems problem. Many optional data are
    included here as well. All other technologies in
    :mod:`osier` inherit from this class.

    Parameters
    ----------
    technology_name : str
        The name identifier of the technology.
    technology_type : str
        The string identifier for the type of technology.
        Two common types are: ["production", "storage"].
    technology_category : str
        The string identifier the the technology category.
        For example: "renewable," "fossil," or "nuclear."
    dispatchable : bool
        Indicates whether the technology can be dispatched by a
        grid operator, or if it produces variable electricity
        that must be used or stored the moment it is produced.
        For example, solar panels and wind turbines are not
        dispatchable, but nuclear and biopower are dispatchable.
        Default value is true.
    renewable : bool
        Indicates whether the technology is considered "renewable."
        Useful for determining if a technology will contribute to
        a renewable portfolio standard (RPS).
    capital_cost : float or :class:`unyt.array.unyt_quantity`
        Specifies the capital cost. If float,
        the default unit is $/MW.
    om_cost_fixed : float or :class:`unyt.array.unyt_quantity`
        Specifies the fixed operating costs.
        If float, the default unit is $/MW.
    om_cost_variable : float, :class:`unyt.array.unyt_quantity`, or array-like
        Specifies the variable operating costs. Users may pass timeseries data.
        However, :class:`pandas.DataFrame` is not supported by this feature.
        If float, the default unit is $/MWh.
    fuel_cost : float, :class:`unyt.array.unyt_quantity`, or array-like
        Specifies the fuel costs. Users may pass timeseries data.
        However, :class:`pandas.DataFrame` is not supported by this feature.
        If float, the default unit is $/MWh.
    fuel_type : str
        Specifies the type of fuel consumed by the technology.
    capacity : float or :class:`unyt.array.unyt_quantity`
        Specifies the technology capacity.
        If float, the default unit is MW
    capacity_factor : Optional, float
        Specifies the 'usable' fraction of a technology's capacity.
        Default is 1.0, i.e. all of the technology's capacity is
        usable all of the time.
    capacity_credit : Optional, float
        Specifies the fraction of a technology's capacity that counts
        towards reliability requirements. Most frequently used for 
        renewable technologies. For example, a solar farm might have 
        a capacity credit of 0.2. This means that in order to meet a 
        capacity requirement of 1 GW, 1.25 GW of solar would need to 
        be installed.
        Default is 1.0, i.e. all of the technology's capacity contributes
        to capacity requirements.
    co2_rate : float or :class:`unyt.array.unyt_quantity`
        Specifies the rate at which carbon dioxide is emitted during operation.
        Generally only applicable for fossil fueled plants.
        If float, the default units are megatonnes per MWh
    lifecycle_co2_rate : float or :class:`unyt.array.unyt_quantity`
        Specifies the rate at which of CO2eq emissions over a typical lifetime. 
        Unless you are reading this in a future where the economy is fully
        decarbonized, all technologies should have a non-zero value for this 
        attribute.
        If float, the default units are megatonnes per MWh
    land_intensity : float or :class:`unyt.array.unyt_quantity`
        The amount of land required per unit capacity. May be either lifecycle
        land use or from direct use. However, consistency between
        technologies is incumbent on the user.
    efficiency : float
        The technology's energy conversion efficiency expressed as
        a fraction. Default is 1.0.
    lifetime : float
        The technology's operational lifetime in years. Default is 25 years.
    default_power_units : str or :class:`unyt.unit_object.Unit`
        An optional parameter, specifies the units
        for power. Default is megawatts [MW].
    default_time_units : str or :class:`unyt.unit_object.Unit`
        An optional parameter, specifies the units
        for time. Default is hours [hr].
    default_mass_units : str or :class:`unyt.unit_object.Unit`
        An optional parameter, specifies the units
        for mass. Default is hours [kg].
    default_energy_units : str or :class:`unyt.unit_object.Unit`
        An optional parameter, specifies the units
        for energy. Default is megawatt-hours [MWh]
        Currently, `default_energy_units` is derived from the
        time and power units.

    Notes
    -----
    Cost values are listed in the docs as [$ / physical unit]. However,
    :class:`osier` does not currently have a currency handler, therefore the
    units are technically [1 / physical unit].

    The :class:`unyt` library may not be able to interpret strings for
    inverse units. For example:

    >>> my_unit = "10 / MW"
    >>> my_unit = unyt_quantity.from_string(my_unit)
    ValueError: Received invalid quantity expression '10/MW'.

    Instead, try the more explicit approach:

    >>> my_unit = "10 MW**-1"
    >>> my_unit = unyt_quantity.from_string(my_unit)
    unyt_quantity(10., '1/MW')

    However, inverse MWh cannot be converted from a string.
    """

    def __init__(self,
                 technology_name,
                 technology_type='production',
                 technology_category='base',
                 dispatchable=True,
                 renewable=False,
                 capital_cost=0.0,
                 om_cost_fixed=0.0,
                 om_cost_variable=0.0,
                 fuel_cost=0.0,
                 fuel_type=None,
                 capacity=0.0,
                 capacity_factor=1.0,
                 capacity_credit=1.0,
                 co2_rate=0.0,
                 lifecycle_co2_rate=0.0,
                 land_intensity=0.0,
                 efficiency=1.0,
                 lifetime=25.0,
                 default_power_units=MW,
                 default_time_units=hr,
                 default_energy_units=None,
                 default_length_units=km,
                 default_volume_units=m**3,
                 default_mass_units=megatonnes) -> None:

        self.technology_name = technology_name
        self.technology_type = technology_type
        self.technology_category = technology_category
        self.dispatchable = dispatchable
        self.renewable = renewable
        self.fuel_type = fuel_type
        self.lifetime = lifetime

        self.unit_power = default_power_units
        self.unit_time = default_time_units
        self.unit_energy = default_energy_units
        self.unit_length = default_length_units
        self.unit_volume = default_volume_units
        self.unit_mass = default_mass_units

        self.capacity = capacity
        self.capacity_factor = capacity_factor
        self.capacity_credit = capacity_credit
        self.efficiency = efficiency
        self.capital_cost = capital_cost
        self.om_cost_fixed = om_cost_fixed
        self.om_cost_variable = om_cost_variable
        self.fuel_cost = fuel_cost
        self.co2_rate = co2_rate
        self.lifecycle_co2_rate = lifecycle_co2_rate
        self.land_intensity = land_intensity

    def __repr__(self) -> str:
        return (f"{self.technology_name}: {self.capacity}")

    def __eq__(self, tech) -> bool:
        """Test technology equality"""
        if ((self.technology_name == tech.technology_name)
                and (self.capacity == tech.capacity)):
            return True
        else:
            return False

    @property
    def unit_power(self):
        return self._unit_power

    @unit_power.setter
    def unit_power(self, value):
        self._unit_power = _validate_unit(value, dimension="power")

    @property
    def unit_time(self):
        return self._unit_time

    @unit_time.setter
    def unit_time(self, value):
        self._unit_time = _validate_unit(value, dimension="time")

    @property
    def unit_mass(self):
        return self._unit_mass

    @unit_mass.setter
    def unit_mass(self, value):
        self._unit_mass = _validate_unit(value, dimension="mass")

    @property
    def unit_length(self):
        return self._unit_length

    @unit_length.setter
    def unit_length(self, value):
        self._unit_length = _validate_unit(value, dimension="length")

    @property
    def unit_area(self):
        return self._unit_length**2

    @unit_area.setter
    def unit_area(self, value):
        self._unit_area = self._unit_length**2

    @property
    def unit_volume(self):
        return self._unit_volume

    @unit_volume.setter
    def unit_volume(self, value):
        self._unit_volume = _validate_unit(value, dimension="volume")

    @property
    def unit_energy(self):
        return self._unit_power * self._unit_time

    @unit_energy.setter
    def unit_energy(self, value):
        self._unit_energy = self._unit_power * self._unit_time

    @property
    def capacity(self):
        return self._capacity.to(self._unit_power)

    @capacity.setter
    def capacity(self, value):
        valid_quantity = _validate_quantity(value, dimension="power")
        self._capacity = valid_quantity.to(self._unit_power)

    @property
    def capital_cost(self):
        return self._capital_cost.to(self._unit_power**-1)

    @capital_cost.setter
    def capital_cost(self, value):
        self._capital_cost = _validate_quantity(
            value, dimension="specific_power")

    @property
    def om_cost_fixed(self):
        return self._om_cost_fixed.to(self._unit_power**-1)

    @om_cost_fixed.setter
    def om_cost_fixed(self, value):
        self._om_cost_fixed = _validate_quantity(
            value, dimension="specific_power")

    @property
    def om_cost_variable(self):
        if isinstance(self._om_cost_variable, _constant_types):
            return self._om_cost_variable.to(self.unit_energy**-1)
        elif isinstance(self._om_cost_variable, _array_types):
            if isinstance(self._om_cost_variable, unyt.unyt_array):
                return self._om_cost_variable.to(self.unit_energy**-1)
            else:
                return np.array(self._om_cost_variable) * \
                    (self.unit_energy**-1)

    @om_cost_variable.setter
    def om_cost_variable(self, value):
        self._om_cost_variable = _validate_quantity(
            value, dimension="specific_energy")

    @property
    def fuel_cost(self):
        if isinstance(self._fuel_cost, _constant_types):
            return self._fuel_cost.to(self.unit_energy**-1)
        elif isinstance(self._fuel_cost, _array_types):
            if isinstance(self._fuel_cost, unyt.unyt_array):
                return self._fuel_cost.to(self.unit_energy**-1)
            else:
                return np.array(self._fuel_cost) * (self.unit_energy**-1)

    @fuel_cost.setter
    def fuel_cost(self, value):
        self._fuel_cost = _validate_quantity(
            value, dimension="specific_energy")

    @property
    def co2_rate(self):
        return self._co2_rate.to(self.unit_mass * self.unit_energy**-1)

    @co2_rate.setter
    def co2_rate(self, value):
        self._co2_rate = _validate_quantity(value, dimension="mass_per_energy")

    @property
    def lifecycle_co2_rate(self):
        return self._lifecycle_co2_rate.to(self.unit_mass * self.unit_energy**-1)

    @lifecycle_co2_rate.setter
    def lifecycle_co2_rate(self, value):
        self._lifecycle_co2_rate = _validate_quantity(value, dimension="mass_per_energy")

    @property
    def land_intensity(self):
        return self._land_intensity.to(self.unit_area * self.unit_power**-1)

    @land_intensity.setter
    def land_intensity(self, value):
        self._land_intensity = _validate_quantity(
            value, dimension="area_per_power")

    @property
    def total_capital_cost(self):
        return self.capacity * self.capital_cost

    @property
    def annual_fixed_cost(self):
        return self.capacity * self.om_cost_fixed

    @property
    def variable_cost(self):
        """
        Combines the fuel and variable operating costs into a total variable cost
        associated with technology usage.

        Notes
        -----
        This function will attempt to merge the two values, even if they have
        different sizes and types. Therefore it is recommended that users
        pass values of the same size and type to prevent unexpected behavior.
        """
        if (isinstance(self.fuel_cost, _constant_types)
                and isinstance(self.om_cost_variable, _constant_types)):
            return self.fuel_cost + self.om_cost_variable
        elif (isinstance(self.fuel_cost, _array_types) and isinstance(self.om_cost_variable, _constant_types)):
            return self.fuel_cost + \
                np.ones(len(self.fuel_cost)) * self.om_cost_variable
        elif (isinstance(self.fuel_cost, _constant_types) and isinstance(self.om_cost_variable, _array_types)):
            return self.fuel_cost * \
                np.ones(len(self.om_cost_variable)) + self.om_cost_variable
        elif (isinstance(self.fuel_cost, _constant_types) and isinstance(self.om_cost_variable, _array_types)):
            return self.fuel_cost * \
                np.ones(len(self.om_cost_variable)) + self.om_cost_variable
        elif (isinstance(self.fuel_cost, _array_types) and isinstance(self.om_cost_variable, _array_types)):
            min_len = min(len(self.fuel_cost), len(self.om_cost_variable))
            return self.fuel_cost[:min_len] + self.om_cost_variable[:min_len]
        else:
            raise TypeError(
                f"Fuel cost has type <{type(self.fuel_cost)}>.\n" +
                f"OM variable cost has type <{type(self.om_cost_variable)}>.\n"
                "One or both of these types are unknown.")

    def variable_cost_ts(self, size):
        """
        Returns the total variable cost as an array of
        length :attr:`size`.

        .. warning::
            The current implementation will only select the
            first N values, where N = `size`. It is recommended
            that users only pass the subset of data they wish
            to use.

        Parameters
        ----------
        size : int
            The number of periods, i.e. length, of the
            time series.

        Returns
        -------
        var_cost_ts : :class:`numpy.ndarray`
            The variable cost time series.
        """
        if isinstance(self.variable_cost, _constant_types):
            var_cost_ts = np.ones(size) * self.variable_cost
            return var_cost_ts

        elif isinstance(self.variable_cost, _array_types):
            try:
                var_cost_ts = self.variable_cost[:size]
                assert len(var_cost_ts) == size
            except AssertionError as e:
                raise AssertionError(
                    f"Variable cost data too short ({len(var_cost_ts)} < {size})")
            return var_cost_ts
        
    def to_dataframe(self, cast_to_string=True):
        """
        Writes all technology attributes to a :class:`pandas.DataFrame` for export
        and manipulation.
        """

        tech_data = OrderedDict()
        tech_data['technology_name'] = [self.technology_name]
        tech_data['technology_category'] = [self.technology_category]
        tech_data['technology_type'] = [self.technology_type]
        tech_data['dispatchable'] = [str(self.dispatchable)]
        tech_data['renewable'] = [str(self.renewable)]
        tech_data['fuel_type'] = [str(self.fuel_type)]

        for key, value in self.__dict__.items():
            if key in tech_data:
                continue
            elif value is None:
                col = key.strip('_')
                tech_data[col] = [str(value)] 
            else:
                if isinstance(value, unyt.unit_object.Unit):
                    continue
                elif isinstance(value, unyt_quantity):
                    col = f"{key.strip('_')} ({value.units})"
                    if cast_to_string:
                        tech_data[col] = ["{:.3g}".format(value.to_value())]
                    else:
                        tech_data[col] = [np.round(value.to_value(),10)]
                elif isinstance(value, (int, float)):
                    col = key.strip('_')
                    if cast_to_string:
                        tech_data[col] = ["{:.3g}".format(value)]
                    else:
                        tech_data[col] = [np.round(value,10)]
                else:
                    continue

        tech_dataframe = pd.DataFrame(tech_data).set_index('technology_name')

        return tech_dataframe
                
            


class RampingTechnology(Technology):
    """
    The :class:`RampingTechnology` class extends the :class:`Technology`
    class by adding ramping attributes that correspond to a technology's
    ability to increase or decrease its power level at a specified rate.

    Parameters
    ----------
    ramp_up_rate : float or :class:`unyt_quantity`
        The rate at which a technology can increase its power, expressed as
        a percentage of its capacity. For example, if `ramp_up_rate` equals 0.5,
        then the technology may ramp up its power level by 50% per unit time.
        The default is 1.0 (i.e. there is no constraint on ramping up).

    ramp_down_rate : float or :class:`unyt_quantity`
        The rate at which a technology can decrease its power, expressed as
        a percentage of its capacity. For example, if `ramp_down_rate` equals 0.5,
        then the technology may ramp down its power level by 50% per unit time.
        The default is 1.0 (i.e. there is no constraint on ramping down).


    Notes
    -----
    It is common for a ramping technology to have different ramp up and ramp
    down rates. Consider a light-water nuclear reactor that can quickly reduce
    its power level by inserting control rods, but must wait much longer to
    increase its power by the same amount due to a build up of neutron
    absorbing isotopes.
    """

    def __init__(
            self,
            technology_type='production',
            technology_category='ramping',
            ramp_up_rate=1.0 * hr**-1,
            ramp_down_rate=1.0 * hr**-1,
            *args,
            **kwargs) -> None:
        self.ramp_up_rate = _validate_quantity(ramp_up_rate,
                                               dimension='specific_time')
        self.ramp_down_rate = _validate_quantity(ramp_down_rate,
                                                 dimension='specific_time')
        super().__init__(technology_type=technology_type,
                         technology_category=technology_category,
                         *args, **kwargs)

    @property
    def ramp_up(self):
        return (
            self.capacity *
            self.ramp_up_rate).to(
            self.unit_power *
            self.unit_time**-1
        )

    @property
    def ramp_down(self):
        return (
            self.capacity *
            self.ramp_down_rate).to(
            self.unit_power *
            self.unit_time**-1
        )


class ThermalTechnology(RampingTechnology):
    """
    The :class:`ThermalTechnology` class extends the :class:`RampingTechnology`
    class by adding a heat rate.

    Parameters
    ----------
    heat_rate : int or float
        The heat rate of a given technology.
    """

    def __init__(
            self,
            heat_rate=None,
            technology_type='production',
            technology_category='thermal',
            *args,
            **kwargs) -> None:
        super().__init__(technology_type=technology_type,
                         technology_category=technology_category,
                         *args, **kwargs)

        self.heat_rate = heat_rate


class StorageTechnology(Technology):
    """
    The :class:`StorageTechnology` extends the :class:`Technology` by
    adding storage parameters.

    Parameters
    ----------
    storage_duration : float or :class:`unyt.array.unyt_quantity`
        The amount of time the battery could discharge continuously when full.
        Used to calculate the storage capacity.
    initial_storage : float or :class:`unyt.array.unyt_quantity`
        The initial stored energy. Cannot exceed :attr:`storage_capacity`.
    """

    def __init__(
            self,
            technology_type='storage',
            storage_duration=0,
            initial_storage=0,
            *args,
            **kwargs) -> None:
        super().__init__(technology_type=technology_type,
                         *args, **kwargs)

        self.storage_duration = storage_duration
        self.initial_storage = initial_storage

    @property
    def storage_duration(self):
        return self._storage_duration

    @storage_duration.setter
    def storage_duration(self, value):
        valid_quantity = _validate_quantity(value, dimension='time')
        self._storage_duration = valid_quantity

    @property
    def storage_capacity(self):
        return self._storage_duration * self._capacity

    @property
    def initial_storage(self):
        return self._initial_storage

    @initial_storage.setter
    def initial_storage(self, value):
        valid_quantity = _validate_quantity(value, dimension='energy')
        try:
            assert valid_quantity <= self.storage_capacity
        except AssertionError:
            raise AssertionError("Initial storage exceeds storage capacity.")

        self._initial_storage = valid_quantity



import pandas as pd
import numpy as np
import pyomo.environ as pe
import pyomo.opt as po
from pyomo.environ import ConcreteModel
from unyt import unyt_array, hr, MW, kW, GW
import itertools as it
from osier import Technology
from osier.technology import _validate_quantity, _validate_unit
from osier.utils import synchronize_units
import warnings
import logging
import time


_freq_opts = {'D': 'day',
              'H': 'hour',
              'S': 'second',
              'T': 'minute'}
LARGE_NUMBER = 1e20
MEDIUM_NUMBER = 1e10
BLACKOUT_COST = 10 * (1 / (kW * hr))  # M$/kWh


curtailment_tech = Technology(technology_name='Curtailment',
                              technology_type='removal',
                              dispatchable=True,
                              capacity=MEDIUM_NUMBER)
reliability_tech = Technology(technology_name='LoadLoss',
                              technology_type='matching',
                              dispatchable=True,
                              fuel_cost=BLACKOUT_COST,
                              capacity=MEDIUM_NUMBER)


class DispatchModel():
    """
    The :class:`DispatchModel` class creates and solves a basic dispatch
    model from the perspective of a "grid operator." The model uses
    `pyomo <https://pyomo.readthedocs.io/en/stable/index.html>`_
    to create and solve a linear programming model.
    The mathematical formulation for this problem is:

    Minimize


    .. math::
        \\text{C}_{\\text{total}} = \\sum_t^T \\sum_u^U
        \\left[c^{\\text{fuel}}_{u,t} + c^{\\text{om,var}}_{u,t}\\right]x_{u,t}

    Such that,

    1. The generation meets demand within a user-specified tolerance
    (undersupply and oversupply)

    .. math::
        \\sum_u^Ux_{u,t} &\\geq \\left(1-\\text{undersupply}\\right)\\text{D}_t
        \\ \\forall \\ t \\in T

        \\sum_u^Ux_{u,t} &\\leq \\left(1+\\text{oversupply}\\right)\\text{D}_t
        \\ \\forall \\ t \\in T

    2. A technology's generation (:math:`x_u`) does not exceed its capacity to
    generate at any time, :math:`t`.

    .. math::
        x_{u,t} \\leq \\textbf{CAP}_{u}\\Delta t \\ \\forall \\ u,t \\in U,T

    3. Technologies may not exceed their ramp up rate,

    .. math::
        \\frac{x_{r,t} - x_{r,t-1}}{\\Delta t} = \\Delta P_{r,t} \\leq
        (\\text{ramp up})\\textbf{CAP}_u\\Delta t \\ \\forall \\ r,t
        \\in R \\subset U, T

    or ramp down rate,

    .. math::
        \\frac{x_{r,t} - x_{r,t-1}}{\\Delta t} = \\Delta P_{r,t} \\leq
        -(\\text{ramp down})\\textbf{CAP}_u\\Delta t \\ \\forall \\ r,t
        \\in R \\subset U, T .

    Parameters
    ----------
    technology_list : list of :class:`osier.Technology`
        The list of :class:`Technology` objects to dispatch -- i.e. decide
        how much energy each technology should produce.
    net_demand : list, :class:`numpy.ndarray`, :class:`unyt.array.unyt_array`, :class:`pandas.DataFrame`
        The remaining energy demand to be fulfilled by the technologies in
        :attr:`technology_list`. The `values` of an object passed as
        `net_demand` are used to create a supply constraint. See
        :attr:`oversupply` and :attr:`undersupply`.
        If a :class:`pandas.DataFrame` is passed, :mod:`osier` will try
        inferring a `time_delta` from the dataframe index. Otherwise, the
        :attr:`time_delta` must be passed or the default is used.
    time_delta : str, :class:`unyt.unyt_quantity`, float, int
        Specifies the amount of time between two time slices. The default is
        one hour. Can be overridden by specifying a unit with the value. For
        example:

        >>> time_delta = "5 minutes"
        >>> from unyt import days
        >>> time_delta = 30*days

        would both work.
    power_units : str, :class:`unyt.unit_object`
        Specifies the units for the power demand. The default is :attr:`MW`.
        Can be overridden by specifying a unit with the value.
    solver : str
        Indicates which solver to use. May require separate installation.
        Accepts: ['cplex', 'cbc']. Other solvers will be added in the future.
    lower_bound : float
        The minimum amount of energy each technology can produce per time
        period. Default is 0.0.
    oversupply : float
        The amount of allowed oversupply as a percentage of demand.
        Default is 0.0 (no oversupply allowed).
    undersupply : float
        The amount of allowed undersupply as a percentage of demand.
        Default is 0.0 (no undersupply allowed).
    verbosity : Optional, int
        Sets the logging level for the simulation. Accepts `logging.LEVEL`
        or integer where LEVEL is {10:DEBUG, 20:INFO, 30:WARNING, 40:ERROR, 50:CRITICAL}.
    curtailment : boolean
        Indicates if the model should enable a curtailment option.
    allow_blackout : boolean
        If True, a "reliability" technology is added to the model that will
        fulfill the mismatch in supply and demand. This reliability technology
        has a variable cost of 1e4 $/MWh. The value must be higher than the
        variable cost of any other technology to prevent a pathological
        preference for blackouts. Default is False.

    Attributes
    ----------
    model : :class:`pyomo.environ.ConcreteModel`
        The :mod:`pyomo` model class that converts python code into a
        set of linear equations.
    objective : float
        The result of the model's objective function. Only instantiated
        after :meth:`DispatchModel.solve()` is called.
    n_timesteps : int
        The number of timesteps in the model.
    upper_bound : float
        The upper bound for all decision variables. Chosen to be equal
        to the maximum capacity of all technologies in :attr:`tech_set`.
    storage_upper_bound : float
        The upper bound for storage decision variables.
    penalty : float
        The penalty applied to the objective function to eliminate
        simultaneous charging and discharging. Users may need to tune
        this parameter. Default is 1e-4.
    model_initialized : bool
        Indicates whether :attr:`DispatchModel.model` has been populated
        with equations yet. This is set to ``True`` after
        :meth:`DispatchModel._write_model_equations()` has been called.
    indices : list of tuples
        The list of tuples representing the product of the
        :attr:`tech_set` and :attr:`time_set` attributes.
    tech_set : list of str
        A list of the unique technology names in the simulation.
    capacity_dict : dict
        A dictionary of {name : capacity} pairs.
    efficiency_dict : dict
        A dictionary of {name : efficiency} pairs.
    time_set : iterable
        The result of ``range(self.n_timesteps)``.
    cost_params : list
        The set of cost parameters for each technology. Corresponds to
        a list of values. Size is equal to the product of the number of
        timesteps and the number of technologies in the model.
    ramp_up_params : list
        The set of ramp_up parameters. Only initialized if there is a
        ramping technology.
    ramp_down_params : list
        The set of ramp_down parameters. Only initialized if there is a
        ramping technology.
    ramping_techs : list
        The subset of :attr:`tech_set` containing ramping technologies.

    Notes
    -----
    1. Technically, :attr:`solver` will accept any solver that :class:`pyomo`
    can use. We only list two solvers because those are the only solvers
    in the :mod:`osier` test suite.

    2. The default value for :attr:`time_delta` in :attr:`__init__` is
    `None`. This is replaced by a setter method in :class:`Dispatch`.

    3. This formulation uses a :attr:`penalty` parameter  to prevent unphysical
    behavior because it preserves the problem's linearity. Some formulations
    may use a binary variable to prevent simultaneous charging and discharing.
    However, this changes the problem to a mixed-integer linear program which
    requires a more sophisticated solver such as ``gurobi``.

    4. :meth:`_write_model_equations` may be called before :meth:`solve` if
    users wish to add their own constraints or parameters to the problem.
    """

    def __init__(self,
                 technology_list,
                 net_demand,
                 time_delta=None,
                 solver='cbc',
                 lower_bound=0.0,
                 oversupply=0.0,
                 undersupply=0.0,
                 verbosity=50,
                 penalty=1e-4,
                 power_units=MW,
                 curtailment=True,
                 allow_blackout=False,
                 **kwargs):
        self.net_demand = net_demand
        self.time_delta = time_delta
        self.lower_bound = lower_bound
        self.oversupply = oversupply
        self.undersupply = undersupply
        self.penalty = penalty
        self.model = ConcreteModel()
        self.solver = solver
        self.results = None
        self.objective = None
        self.model_initialized = False
        self.verbosity = verbosity
        if self.verbosity > 10:
            self.verbose = False
        else: 
            self.verbose = True
        self.curtailment = curtailment
        self.allow_blackout = allow_blackout

        if isinstance(net_demand, unyt_array):
            self.power_units = net_demand.units
        else:
            self.power_units = power_units

        if ((self.curtailment) and (self.allow_blackout)):
            sync_list = technology_list + [curtailment_tech, reliability_tech]
        elif self.curtailment:
            sync_list = technology_list + [curtailment_tech]
        elif self.allow_blackout:
            sync_list = technology_list + [reliability_tech]
        else:
            sync_list = technology_list

        self.technology_list = synchronize_units(
            sync_list,
            unit_power=self.power_units,
            unit_time=self.time_delta.units)

        logging.basicConfig(level=verbosity, format='%(message)s')


    @property
    def time_delta(self):
        return self._time_delta

    @time_delta.setter
    def time_delta(self, value):
        if value:
            valid_quantity = _validate_quantity(value, dimension='time')
            self._time_delta = valid_quantity
        else:
            if isinstance(self.net_demand, pd.DataFrame):
                try:
                    freq_list = list(self.net_demand.index.inferred_freq)
                    freq_key = freq_list[-1]
                    try:
                        value = float(freq_list[0])
                    except ValueError:
                        warnings.warn((f"Could not convert value "
                                       f"{freq_list[0]} to float. "
                                       "Setting to 1.0."),
                                      UserWarning)
                        value = 1.0
                    self._time_delta = _validate_quantity(
                        f"{value} {_freq_opts[freq_key]}", dimension='time')
                except KeyError:
                    warnings.warn(
                        (f"Could not infer time delta with freq {freq_key} "
                         "from pandas dataframe. Setting delta to 1 hour."),
                        UserWarning)
                    self._time_delta = 1 * hr
            else:
                self._time_delta = 1 * hr

    @property
    def power_units(self):
        return self._power_units

    @power_units.setter
    def power_units(self, value):
        if value:
            valid_quantity = _validate_unit(value, dimension='power')
            self._power_units = valid_quantity
        else:
            warnings.warn(f"Could not infer power units. Unit set to MW.")
            self._power_units = MW

    @property
    def n_timesteps(self):
        return len(self.net_demand)

    @property
    def tech_set(self):
        tech_names = [t.technology_name for t in self.technology_list]
        return tech_names

    @property
    def capacity_dict(self):
        capacity_set = unyt_array([t.capacity for t in self.technology_list])
        return dict(zip(self.tech_set, capacity_set))

    @property
    def efficiency_dict(self):
        capacity_set = [t.efficiency for t in self.technology_list]
        return dict(zip(self.tech_set, capacity_set))

    @property
    def time_set(self):
        return range(self.n_timesteps)

    @property
    def indices(self):
        return list(it.product(self.tech_set, self.time_set))

    @property
    def cost_params(self):
        v_costs = np.array([
            (t.variable_cost_ts(self.n_timesteps))
            for t in self.technology_list
        ]).flatten()
        return dict(zip(self.indices, v_costs))

    @property
    def storage_techs(self):
        return [t.technology_name
                for t in self.technology_list
                if hasattr(t, "storage_capacity")]

    @property
    def storage_upper_bound(self):
        caps = unyt_array([t.storage_capacity
                           for t in self.technology_list
                           if hasattr(t, "storage_capacity")])
        return caps.max().to_value()

    @property
    def max_storage_params(self):
        storage_dict = {t.technology_name: t.storage_capacity.to_value()
                        for t in self.technology_list
                        if hasattr(t, "storage_capacity")}
        return storage_dict

    @property
    def initial_storage_params(self):
        storage_dict = {t.technology_name: t.initial_storage.to_value()
                        for t in self.technology_list
                        if hasattr(t, "initial_storage")}
        return storage_dict

    @property
    def ramping_techs(self):
        return [t.technology_name
                for t in self.technology_list
                if hasattr(t, 'ramp_up')]

    @property
    def ramp_up_params(self):
        rates_dict = {
            t.technology_name: (
                t.ramp_up *
                self.time_delta).to_value() for t in self.technology_list
            if hasattr(t, 'ramp_up')}
        return rates_dict

    @property
    def ramp_down_params(self):
        rates_dict = {
            t.technology_name: (
                t.ramp_down *
                self.time_delta).to_value() for t in self.technology_list
            if hasattr(t, 'ramp_up')}
        return rates_dict

    @property
    def upper_bound(self):
        caps = unyt_array([t.capacity for t in self.technology_list])
        return caps.max().to_value()

    def _create_model_indices(self):
        self.model.Generators = pe.Set(initialize=self.tech_set, ordered=True)
        self.model.Time = pe.Set(initialize=self.time_set, ordered=True)
        if len(self.ramping_techs) > 0:
            self.model.RampingTechs = pe.Set(initialize=self.ramping_techs,
                                             ordered=True,
                                             within=self.model.Generators)
        if len(self.storage_techs) > 0:
            self.model.StorageTech = pe.Set(initialize=self.storage_techs,
                                            ordered=True,
                                            within=self.model.Generators)

    def _create_demand_param(self):
        self.model.Demand = pe.Param(self.model.Time, initialize=dict(
            zip(self.model.Time, np.array(self.net_demand))))

    def _create_cost_param(self):
        self.model.VariableCost = pe.Param(
            self.model.Generators,
            self.model.Time,
            initialize=self.cost_params)

    def _create_ramping_params(self):
        self.model.ramp_up = pe.Param(
            self.model.RampingTechs, initialize=self.ramp_up_params)
        self.model.ramp_down = pe.Param(
            self.model.RampingTechs, initialize=self.ramp_down_params)

    def _create_max_storage_params(self):
        self.model.storage_capacity = pe.Param(
            self.model.StorageTech, initialize=self.max_storage_params
        )

    def _create_init_storage_params(self):
        self.model.initial_storage = pe.Param(
            self.model.StorageTech, initialize=self.initial_storage_params
        )

    def _create_model_variables(self):
        self.model.x = pe.Var(self.model.Generators, self.model.Time,
                              domain=pe.NonNegativeReals,
                              bounds=(self.lower_bound, self.upper_bound))

        if len(self.storage_techs) > 0:
            self.model.storage_level = pe.Var(
                self.model.StorageTech,
                self.model.Time,
                domain=pe.NonNegativeReals,
                bounds=(
                    self.lower_bound,
                    self.storage_upper_bound))
            self.model.charge = pe.Var(self.model.StorageTech, self.model.Time,
                                       domain=pe.NonNegativeReals,
                                       bounds=(self.lower_bound,
                                               self.upper_bound))

    def _objective_function(self):
        expr = sum(self.model.VariableCost[g, t] * self.model.x[g, t]
                   for g in self.model.Generators for t in self.model.Time)
        if len(self.storage_techs) > 0:
            expr += sum(self.model.x[s, t] + self.model.charge[s, t]
                        for s in self.model.StorageTech
                        for t in self.model.Time) * self.penalty
        self.model.objective = pe.Objective(sense=pe.minimize, expr=expr)

    def _supply_constraints(self):
        self.model.oversupply = pe.ConstraintList()
        self.model.undersupply = pe.ConstraintList()
        for t in self.model.Time:
            generation = sum(self.model.x[g, t] for g in self.model.Generators
                             if g != 'Curtailment')
            if self.curtailment:
                generation -= self.model.x['Curtailment', t]
            if len(self.storage_techs) > 0:
                generation -= sum(self.model.charge[s, t]
                                  for s in self.model.StorageTech)
            over_demand = self.model.Demand[t] * (1 + self.oversupply)
            under_demand = self.model.Demand[t] * (1 - self.undersupply)
            self.model.oversupply.add(generation <= over_demand)
            self.model.undersupply.add(generation >= under_demand)

    def _generation_constraint(self):
        self.model.gen_limit = pe.ConstraintList()
        for g in self.model.Generators:
            unit_capacity = (
                self.capacity_dict[g] *
                self.time_delta).to_value()

            for t in self.model.Time:
                unit_generation = self.model.x[g, t]
                self.model.gen_limit.add(unit_generation <= unit_capacity)

    def _ramping_constraints(self):
        self.model.ramp_up_limit = pe.ConstraintList()
        self.model.ramp_down_limit = pe.ConstraintList()

        for r in self.model.RampingTechs:
            ramp_up = self.model.ramp_up[r]
            ramp_down = self.model.ramp_down[r]
            for t in self.model.Time:
                if t != self.model.Time.first():
                    t_prev = self.model.Time.prev(t)
                    previous_gen = self.model.x[r, t_prev]
                    current_gen = self.model.x[r, t]
                    delta_power = (current_gen - previous_gen) / \
                        self.time_delta.to_value()
                    self.model.ramp_up_limit.add(delta_power <= ramp_up)
                    self.model.ramp_down_limit.add(delta_power >= -ramp_down)

    def _storage_constraints(self):
        self.model.discharge_limit = pe.ConstraintList()
        self.model.charge_limit = pe.ConstraintList()
        self.model.charge_rate_limit = pe.ConstraintList()
        self.model.storage_limit = pe.ConstraintList()
        self.model.set_storage = pe.ConstraintList()
        for s in self.model.StorageTech:
            efficiency = self.efficiency_dict[s]
            storage_cap = self.model.storage_capacity[s]
            unit_capacity = (
                self.capacity_dict[s] *
                self.time_delta).to_value()
            initial_storage = self.model.initial_storage[s]
            for t in self.model.Time:
                self.model.charge_rate_limit.add(
                    self.model.charge[s, t] <= unit_capacity)
                if t == self.model.Time.first():
                    self.model.set_storage.add(self.model.storage_level[s, t]
                                               == initial_storage)
                    self.model.discharge_limit.add(
                        self.model.x[s, t] <= initial_storage)
                    self.model.charge_limit.add(self.model.charge[s, t]
                                                <= storage_cap
                                                - initial_storage)
                else:
                    t_prev = self.model.Time.prev(t)
                    previous_storage = self.model.storage_level[s, t_prev]
                    current_discharge = self.model.x[s, t]
                    current_charge = self.model.charge[s, t]
                    self.model.set_storage.add(
                        self.model.storage_level[s, t]
                        == previous_storage
                        + np.sqrt(efficiency) * current_charge
                        - np.sqrt(efficiency) * current_discharge
                    )
                    self.model.charge_limit.add(
                        np.sqrt(efficiency) * current_charge
                        <= storage_cap - previous_storage)
                    self.model.discharge_limit.add(
                        current_discharge <= previous_storage)
                self.model.storage_limit.add(
                    self.model.storage_level[s, t] <= storage_cap)

    def _write_model_equations(self):

        self._create_model_indices()
        self._create_demand_param()
        self._create_cost_param()
        self._create_model_variables()
        self._objective_function()
        self._supply_constraints()
        self._generation_constraint()
        if len(self.ramping_techs) > 0:
            self._create_ramping_params()
            self._ramping_constraints()
        if len(self.storage_techs) > 0:
            self._create_init_storage_params()
            self._create_max_storage_params()
            self._storage_constraints()
        self.model_initialized = True

    def _format_results(self):
        df = pd.DataFrame(index=self.model.Time)
        for g in self.model.Generators:
            if g == 'Curtailment':
                df[g] = [-1 * self.model.x[g, t].value for t in self.model.Time]
            else:
                df[g] = [self.model.x[g, t].value for t in self.model.Time]

        if len(self.storage_techs) > 0:
            for s in self.model.StorageTech:
                df[f"{s}_charge"] = [-1 * self.model.charge[s, t].value
                                     for t in self.model.Time]
                df[f"{s}_level"] = [self.model.storage_level[s, t].value
                                    for t in self.model.Time]

        df["Cost"] = np.array(sum(self.model.VariableCost[g, t] * self.model.x[g, t].value
                                  for g in self.model.Generators) for t in self.model.Time)

        return df

    def solve(self, solver=None):
        """
        Executes the model solve. Model equations are written at the
        time the method is called.

        Parameters
        ----------
        solver : str
            Indicates which solver to use. If no solver is specified,
            the default :attr:`DispatchModel.solver` attribute is used.
            Default is ['cplex'].
        """
        if not self.model_initialized:
            self._write_model_equations()

        if solver:
            optimizer = po.SolverFactory(solver)
        else:
            optimizer = po.SolverFactory(self.solver)

        results = optimizer.solve(self.model, tee=self.verbose)
        try:
            self.objective = self.model.objective()
        except ValueError:
            if self.verbosity <= 30:
                warnings.warn(
                    f"Infeasible or no solution. Objective set to {LARGE_NUMBER}")
            self.objective = LARGE_NUMBER

        try:
            self.results = self._format_results()
        except TypeError:
            self.results = None



from osier.equations import (annualized_capital_cost, annualized_fixed_cost, total_cost, 
                             annual_emission, objective_from_energy, objective_from_capacity)
from osier import Technology, DispatchModel
import numpy as np
import pytest
import sys
import functools

if "win32" in sys.platform:
    solver = 'cbc'
elif "linux" in sys.platform:
    solver = "cbc"
else:
    solver = "cbc"

TOL = 1e-5
N_HOURS = 24
N_DAYS = 2
N = N_HOURS * N_DAYS


@pytest.fixture(scope="session")
def technology_set_1():
    """
    This fixture creates technologies directly from
    the :class:`Technology` class.
    """
    nuclear = Technology(technology_name='Nuclear',
                         technology_type='production',
                         capacity=5,
                         capital_cost=6,
                         om_cost_variable=20,
                         om_cost_fixed=50,
                         fuel_cost=5
                         )
    natural_gas = Technology(technology_name='NaturalGas',
                             technology_type='production',
                             capacity=5,
                             capital_cost=1,
                             om_cost_variable=12,
                             om_cost_fixed=30,
                             fuel_cost=20
                             )

    nuclear.co2_rate = 1.2e-5
    natural_gas.co2_rate = 4.9e-4

    return [nuclear, natural_gas]

@pytest.fixture
def net_demand():

    phase_shift = 0
    base_load = 1.5
    hours = np.linspace(0, N, N)
    demand = (np.sin((hours * np.pi / N_HOURS * 2 + phase_shift))
              * -1 + np.ones(N) * (base_load + 1))

    return demand


def test_annualized_capital_cost(technology_set_1):
    """
    Tests the annualized capital cost is calculated correctly.
    """
    nuclear, natural_gas = technology_set_1

    expected = (nuclear.total_capital_cost / nuclear.lifetime) \
               + (natural_gas.total_capital_cost / natural_gas.lifetime)
    actual = annualized_capital_cost(technology_list=technology_set_1)

    assert expected == pytest.approx(actual)


def test_annualized_fixed_cost(technology_set_1):
    """
    Tests the annualized fixed cost is calculated correctly.
    """
    nuclear, natural_gas = technology_set_1

    expected = (nuclear.annual_fixed_cost) + (natural_gas.annual_fixed_cost)
    actual = annualized_fixed_cost(technology_list=technology_set_1)

    assert expected == pytest.approx(actual)


def test_total_cost(technology_set_1, net_demand):
    """
    Tests that :func:`total_cost` produces expected results.
    """
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    model.solve()

    expected = annualized_fixed_cost(technology_set_1) \
                + annualized_capital_cost(technology_set_1) \
                + model.objective
    actual = total_cost(technology_set_1, model)

    assert expected == pytest.approx(actual)


def test_annual_co2(technology_set_1, net_demand):
    """
    Tests that :func:`annual_co2` produces expected results.
    """
    nuclear, natural_gas = technology_set_1
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    model.solve()

    expected = (model.results["Nuclear"].sum() * nuclear.co2_rate) \
                + (model.results["NaturalGas"].sum() * natural_gas.co2_rate)
    actual = annual_emission(technology_set_1, model, emission='co2_rate')

    assert expected == pytest.approx(actual)


def test_objective_from_capacity(technology_set_1, net_demand):
    """
    Tests that :func:`objective_from_capacity` produces expected results.
    """
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    model.solve()

    func = functools.partial(objective_from_capacity, attribute='om_cost_fixed')
    expected = annualized_fixed_cost(technology_set_1, model)
    actual = func(technology_list=technology_set_1, 
                  solved_dispatch_model=model)

    assert expected == pytest.approx(actual)


def test_objective_from_energy(technology_set_1, net_demand):
    """
    Tests that :func:`objective_from_energy` produces expected results.
    """
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    model.solve()

    func = functools.partial(objective_from_energy, attribute='co2_rate')
    expected = annual_emission(technology_set_1, model, emission='co2_rate')
    actual = func(technology_list=technology_set_1, 
                  solved_dispatch_model=model)

    assert expected == pytest.approx(actual)


from osier import DispatchModel
from osier import Technology, ThermalTechnology, StorageTechnology, RampingTechnology
from unyt import unyt_array
import unyt
import numpy as np
import pandas as pd
import pytest
import sys

if "win32" in sys.platform:
    solver = 'cbc'
elif "linux" in sys.platform:
    solver = "cbc"
else:
    solver = "cbc"

TOL = 1e-5
N_HOURS = 24
N_DAYS = 2
N = N_HOURS * N_DAYS


@pytest.fixture(scope="session")
def technology_set_1():
    """
    This fixture creates technologies directly from
    the :class:`Technology` class.
    """
    nuclear = Technology(technology_name='Nuclear',
                         technology_type='production',
                         capacity=5,
                         capital_cost=6,
                         om_cost_variable=20,
                         om_cost_fixed=50,
                         fuel_cost=5
                         )
    natural_gas = Technology(technology_name='NaturalGas',
                             technology_type='production',
                             capacity=5,
                             capital_cost=1,
                             om_cost_variable=12,
                             om_cost_fixed=30,
                             fuel_cost=20
                             )

    return [nuclear, natural_gas]


@pytest.fixture
def technology_set_2():
    """
    This fixture creates technologies from
    the :class:`ThermalTechnology` subclass.
    """
    nuclear = ThermalTechnology(technology_name='Nuclear',
                                capacity=5,
                                capital_cost=6,
                                om_cost_variable=20,
                                om_cost_fixed=50,
                                fuel_cost=5,
                                ramp_up_rate=0.0,
                                ramp_down_rate=0.0,
                                )
    natural_gas = RampingTechnology(technology_name='NaturalGas',
                                    capacity=5,
                                    capital_cost=1,
                                    om_cost_variable=12,
                                    om_cost_fixed=30,
                                    fuel_cost=20,
                                    ramp_up_rate=0.9,
                                    ramp_down_rate=0.9,
                                    )

    return [nuclear, natural_gas]


@pytest.fixture
def technology_set_3():
    """
    This fixture creates technologies from
    the :class:`ThermalTechnology` subclass.
    """
    nuclear = ThermalTechnology(technology_name='Nuclear',
                                capacity=5,
                                capital_cost=6,
                                om_cost_variable=20,
                                om_cost_fixed=50,
                                fuel_cost=5,
                                ramp_up_rate=0.05,
                                ramp_down_rate=0.05,
                                )
    natural_gas = ThermalTechnology(technology_name='NaturalGas',
                                    capacity=5,
                                    capital_cost=1,
                                    om_cost_variable=12,
                                    om_cost_fixed=30,
                                    fuel_cost=20,
                                    ramp_up_rate=0.9,
                                    ramp_down_rate=0.9,
                                    )

    return [nuclear, natural_gas]


@pytest.fixture
def technology_set_4():
    """
    This fixture uses creates technologies from
    the :class:`ThermalTechnology` and :class`StorageTechnology`
    subclasses.
    """
    nuclear = ThermalTechnology(technology_name='Nuclear',
                                capacity=5,
                                capital_cost=6,
                                om_cost_variable=20,
                                om_cost_fixed=50,
                                fuel_cost=5,
                                ramp_up_rate=0.0,
                                ramp_down_rate=0.0,
                                )
    battery = StorageTechnology(technology_name='Battery',
                                capacity=5,
                                capital_cost=1,
                                om_cost_variable=0,
                                om_cost_fixed=15,
                                fuel_cost=0,
                                storage_duration=13,
                                efficiency=1.0,
                                initial_storage=24
                                )

    return [nuclear, battery]


@pytest.fixture
def net_demand():

    phase_shift = 0
    base_load = 1.5
    hours = np.linspace(0, N, N)
    demand = (np.sin((hours * np.pi / N_HOURS * 2 + phase_shift))
              * -1 + np.ones(N) * (base_load + 1))

    return demand


def test_dispatch_model_initialize(technology_set_1, net_demand):
    """
    Tests that the dispatch model is properly initialized.
    """
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    assert model.technology_list == technology_set_1
    assert model.tech_set == [t.technology_name for t in technology_set_1]
    assert model.solver == solver
    assert len(model.capacity_dict) == len(technology_set_1)
    assert len(model.indices) == len(net_demand) * len(technology_set_1)
    assert model.time_delta == 1 * unyt.hour


@pytest.mark.filterwarnings("ignore")
def test_dispatch_model_time_delta(technology_set_1, net_demand):
    """
    Tests that the model properly initializes the time delta attribute.
    """
    t1 = pd.date_range('1/1/2022', '6/1/2022', freq='2D')[:N]
    t2 = pd.date_range('1/1/2022', '3/1/2070', freq='Y')[:N]
    df1 = pd.DataFrame({'data': net_demand}, index=t1)
    df2 = pd.DataFrame({'data': net_demand}, index=t2)

    model1 = DispatchModel(technology_set_1,
                           net_demand=df1,
                           solver=solver)
    model2 = DispatchModel(technology_set_1,
                           net_demand=df2,
                           solver=solver)
    model3 = DispatchModel(technology_set_1,
                           net_demand=net_demand,
                           solver=solver)

    assert model1.time_delta == 2 * unyt.day
    assert model2.time_delta == 1 * unyt.hour

    model3.time_delta = "2 hr"
    assert model3.time_delta == 2 * unyt.hour


def test_dispatch_model_solve_case1(technology_set_1, net_demand):
    """
    Tests that the dispatch model produces expected results. Where all
    the technologies are simply :class:`Technology` objects. The model
    should always choose the cheapest technology, as long as it has
    enough capacity to meet the demand.
    """
    model = DispatchModel(technology_set_1,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False,
                          allow_blackout=False)
    model.solve()
    cheapest_tech = unyt_array(
        [t.variable_cost for t in technology_set_1]).min()
    expected_result = cheapest_tech * net_demand.sum()

    assert model.objective == pytest.approx(expected_result, rel=TOL)
    assert model.results['Nuclear'].sum(
    ) == pytest.approx(net_demand.sum(), TOL)
    assert model.results['NaturalGas'].sum() == pytest.approx(0.0, rel=TOL)


def test_dispatch_model_solve_case2(technology_set_2, net_demand):
    """
    Tests that the dispatch model produces expected results. The technologies
    are :class:`ThermalTechnology` objects. In this case, the `Nuclear`
    technology is not allowed to change its power level at all. Therefore
    we expect Nuclear to fulfill a "baseload" role, with all other demand met
    by natural gas.
    """
    nuclear, natgas = technology_set_2
    model = DispatchModel(technology_set_2,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False)
    model.solve()
    expected_nuclear = net_demand.min() * np.ones(N)
    expected_natgas = net_demand - expected_nuclear
    expected_result = (expected_nuclear * nuclear.variable_cost
                       + expected_natgas * natgas.variable_cost).sum()
    assert model.objective == pytest.approx(expected_result, rel=TOL)


def test_dispatch_model_solve_case3(technology_set_3, net_demand):
    """
    Tests that a dispatch model with ramping constraints behaves
    as expected.
    """
    nuclear, natgas = technology_set_3
    model = DispatchModel(technology_set_3,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False)
    model.solve()
    max_power_delta = ((model.results.Nuclear.diff())
                       / nuclear.capacity.to_value()).max()
    min_power_delta = ((model.results.Nuclear.diff())
                       / nuclear.capacity.to_value()).min()
    assert max_power_delta == pytest.approx(
        nuclear.ramp_up_rate.to_value(), abs=TOL)
    assert min_power_delta == pytest.approx(
        -nuclear.ramp_down_rate.to_value(), abs=TOL)


def test_dispatch_model_solve_case4(technology_set_4, net_demand):
    """
    Tests that storage constraints behave as expected.
    """

    model = DispatchModel(technology_set_4,
                          net_demand=net_demand,
                          solver=solver,
                          curtailment=False)
    model.solve()
    total_gen = model.results[['Nuclear',
                               'Battery',
                               'Battery_charge']].sum().sum()
    binary_charging = np.dot(model.results.Battery,
                             model.results.Battery_charge)
    assert (total_gen - net_demand.sum()) == pytest.approx(0, abs=TOL)
    assert binary_charging == pytest.approx(0, abs=TOL)

def test_dispatch_model_solve_case5(technology_set_4, net_demand):
    """
    Tests that the curtailment technology behaves as expected.
    """

    nuclear, battery = technology_set_4
    model = DispatchModel([nuclear],
                          net_demand=net_demand,
                          solver=solver)
    model.solve()
    total_gen = model.results[['Nuclear',
                               'Curtailment']].sum().sum()
    assert (total_gen - net_demand.sum()) == pytest.approx(0, abs=TOL)


def test_dispatch_model_solve_case6(technology_set_4, net_demand):
    """
    Tests that the reliability technology behaves as expected.
    """

    nuclear, battery = technology_set_4
    nuclear.capacity = 2
    model = DispatchModel([nuclear],
                          net_demand=net_demand,
                          solver=solver,
                          allow_blackout=True)
    model.solve()
    total_gen = model.results[['Nuclear',
                               'Curtailment',
                               'LoadLoss']].sum().sum()
    assert (total_gen - net_demand.sum()) == pytest.approx(0, abs=TOL)
